/*=========================================================================*
 *	File    : bigint.c
 *
 *	Author  : Pallav Gupta
 *	Created : Sat Sep 27 21:59:00 2003
 *
 * This source code can be redistributed under the GNU Free Public Licence.
 *=========================================================================*/

/* $Id: bigint.c,v 1.3 2003/12/31 00:47:05 pgupta Exp $ */

#include "sis.h"
#include "bigint.h"

/* Local functions used internally */
static ELEMENT ascii_to_hex(ELEMENT);
static int bigint_setbits(BIGINT *, INDEX, ELEMENT, ELEMENT, ELEMENT);
static int bigint_getbits(BIGINT *, INDEX, ELEMENT, ELEMENT);

/* allocate and return a bigint */
BIGINT *bigint_new(void)
{
  BIGINT *a;
  a = ALLOC(BIGINT, 1);
  bigint_null(a);
  return a;
}

/* free a bigint, returns BIGINT_OK on success, BIGINT_ERROR on failure */
int bigint_free(BIGINT *a)
{
  CHECK_CONDITION(a == NIL(BIGINT), BIGINT_ERROR);
  FREE(a);
  return BIGINT_OK;
}

/* free an array of bigints */
int bigint_array_free(BIGINT **a, INDEX size) 
{
  INDEX i;
  CHECK_CONDITION(a == NIL(BIGINT *), BIGINT_ERROR);
  for (i = 0; i < size; i++) { bigint_free(a[i]); }
  return BIGINT_OK;
}

/* set a bigint to 0, returns BIGINT_OK */
int bigint_null(BIGINT *a)
{ 
  INDEX i;
  for(i=MAXWORDINDEX; i >=0; i--) { a->data[i] = 0; }
  return BIGINT_OK;
}

/* print a bigint in hexadecimal format, returns BIGINT_OK on success 
   assumes a point to a valid bigint
*/
int bigint_print(BIGINT *a, FILE *fp)
{
  INDEX i;
  for(i=MAXWORDINDEX; i >=0; i--) { fprintf(fp, "%0*x\n", WORDSIZE/4, a->data[i]); } 
  fprintf(fp,"\n");
  return BIGINT_OK;
}

/* returns a duplicate copy of bigint, assumes a points to a valid bigint */
BIGINT *bigint_copy(BIGINT *a) 
{
  BIGINT *b;
  INDEX i;
  b = bigint_new();
  for(i=MAXWORDINDEX; i >=0; i--) { b->data[i] = a->data[i]; }
  return b;
}

/* compare a and b upto maxindex (starting at 0), return 0 if equal, otherwise return index of word at which they differ
 return BIGINT_ERROR on problem */
int bigint_compare(BIGINT *a, BIGINT *b, INDEX maxindex)
{
  INDEX i;
  
  CHECK_CONDITION((a == NIL(BIGINT) || b == NIL(BIGINT)), BIGINT_ERROR);
  CHECK_CONDITION(maxindex > MAXBITINDEX || maxindex < 0, BIGINT_ERROR);
  
  for (i = maxindex; i >= 0; i--)
    if (bigint_get_bit(a, i) != bigint_get_bit(b, i))
      return BIGINT_DIFF;
  
  return BIGINT_OK;
}

/* extract the i^th bit in a, returns BIGINT_ERROR on error or the selected bit on success 
   assumes a points to a valid bigint */
int bigint_get_bit(BIGINT *a, INDEX i)
{
  CHECK_CONDITION(i > MAXBITINDEX || i < 0, BIGINT_ERROR);
  return (a->data[i/WORDSIZE] >> (i%WORDSIZE)) & BIT_ZERO;
}

/* set the i^th bit in a to v, return BIGINT_ERROR on error, BIGINT_OK on success 
   assumes a points to a valid bigint */
int bigint_set_bit(BIGINT *a, INDEX i, ELEMENT v)
{
  CHECK_CONDITION(i > MAXBITINDEX || i < 0, BIGINT_ERROR);
  a->data[i/WORDSIZE] |= (v << (i%WORDSIZE));
  return BIGINT_OK;
}

/* returns data[i], BIGINT_ERROR on error */
int bigint_get_word(BIGINT *a, INDEX i)
{
  CHECK_CONDITION(i > MAXWORDINDEX || i < 0, BIGINT_ERROR);
  return a->data[i];
}

/* returns bits 16-31, BIGINT_ERROR on error */
int bigint_get_upper_halfword(BIGINT *a, INDEX i)
{ return bigint_getbits(a, i, UPPER_WORD, WORDSIZE/2); }
/* returns bits 0-15, BIGINT_ERROR on error */
int bigint_get_lower_halfword(BIGINT *a, INDEX i)
{ return bigint_getbits(a, i, LOWER_WORD, 0); }
/* returns bits 0-7, BIGINT_ERROR on error */
int bigint_get_lower_lowbyte(BIGINT *a, INDEX i)
{ return bigint_getbits(a, i, LOWER_LOWBYTE, 0); }
/* returns bits 8-15, BIGINT_ERROR on error */
int bigint_get_lower_highbyte(BIGINT *a, INDEX i)
{ return bigint_getbits(a, i, LOWER_HIGHBYTE, WORDSIZE/4); }
/* returns bits 16-23, BIGINT_ERROR on error */
int bigint_get_upper_lowbyte(BIGINT *a, INDEX i)
{ return bigint_getbits(a, i, UPPER_LOWBYTE, WORDSIZE/2); }
/* returns bits 24-31, BIGINT_ERROR on error */
int bigint_get_upper_highbyte(BIGINT *a, INDEX i)
{ return bigint_getbits(a, i, UPPER_HIGHBYTE, (WORDSIZE/4)*3); }

/* sets the required word (32 bits) of a bigint to value v, return BIGINT_ERROR on error, BIGINT_OK on success 
   assumes a points to a valid bigint */
int bigint_set_word(BIGINT *a, INDEX i, ELEMENT v) 
{
  CHECK_CONDITION(i > MAXWORDINDEX || i < 0, BIGINT_ERROR);
  a->data[i] = v;
  return BIGINT_OK;
}

/* sets bits 16-31 of word to value v,  return BIGINT_ERROR on error, BIGINT_OK on success 
   assumes a points to a valid bigint */ 
int bigint_set_upper_halfword(BIGINT *a, INDEX i, ELEMENT v)
{
  CHECK_CONDITION(i > MAXWORDINDEX || i < 0, BIGINT_ERROR);
  a->data[i] |= ((v & LOWER_WORD) << (WORDSIZE/2));
  return BIGINT_OK;
}

/* sets bits 0-15 of word to value v,  return BIGINT_ERROR on error, BIGINT_OK on success 
   assumes a points to a valid bigint */ 
int bigint_set_lower_halfword(BIGINT *a, INDEX i, ELEMENT v)
{
  CHECK_CONDITION(i > MAXWORDINDEX || i < 0, BIGINT_ERROR);
  a->data[i] |= (v & LOWER_WORD);
  return BIGINT_OK;
}

/* sets bits 0-7 of word to value v,  return BIGINT_ERROR on error, BIGINT_OK on success 
   assumes a points to a valid bigint */ 
int bigint_set_lower_lowbyte(BIGINT *a, INDEX i, ELEMENT v)
{ return bigint_setbits(a, i, v, LOWER_LOWBYTE, 0); }

/* sets bits 8-15 of word to value v,  return BIGINT_ERROR on error, BIGINT_OK on success 
   assumes a points to a valid bigint */ 
int bigint_set_lower_highbyte(BIGINT *a, INDEX i, ELEMENT v)
{ return bigint_setbits(a, i, v, LOWER_HIGHBYTE, WORDSIZE/4); }

/* sets bits 16-23 of word to value v,  return BIGINT_ERROR on error, BIGINT_OK on success 
   assumes a points to a valid bigint */ 
int bigint_set_upper_lowbyte(BIGINT *a, INDEX i, ELEMENT v)
{ return bigint_setbits(a, i, v, UPPER_LOWBYTE, WORDSIZE/2); }

/* sets bits 24-31 of word to value v, return BIGINT_ERROR on error, BIGINT_OK on success 
   assumes a points to a valid bigint */ 
int bigint_set_upper_highbyte(BIGINT *a, INDEX i, ELEMENT v)
{ return bigint_setbits(a, i, v, UPPER_HIGHBYTE, (WORDSIZE/4)*3); }

/* converts an ascii (in binary/hexadecimal) to a bigint, returns a NULL on error, BIGINT on success
   
   Examples: 
   1) to convert 010101 to bigint, you'd say ascii_to_bigint("010101", 6, 2);
   2) to convert ABCD01 to bigint, you'd say ascii_to_bigint("ABCD01", 6, 16);
*/
BIGINT *ascii_to_bigint(char *str, ELEMENT length, ELEMENT base)
{
  ELEMENT ch;
  INDEX i, j, k;
  BIGINT *a;
  
  CHECK_CONDITION(base != BINARY && base != HEXADECIMAL, NULL);
  a = bigint_new();
  
  /* binary */
  if (base == BINARY) {
    for (i = 0; i < length; i++) {
      ch = (*str++ == '0') ? 0 : 1;
      bigint_set_bit(a, length-i-1, ch);
    }
  }
  /* hexadecimal */
  else {
    for (i = 0; i < length; i++) {
      if ((ch = ascii_to_hex(*str++)) != BIGINT_ERROR) {
	for (j = 3, k = 0; j >= 0; j--)
	  bigint_set_bit(a, HEXBITS*length - HEXBITS*i - ++k, ((ch >> j) & BIT_ZERO));
      }
      else
	return NIL(BIGINT);
    }
  }
  
  return a;
}

/* convert an array of vectors (binary/hexadecimal) into parallel bigint form,
   returns NULL on error, pointer to an array of BIGINT on success
   an array of pointers is returned in a 

   Example:
   
   0000
   0001
   0010
   0011
   0100
   
   ascii_to_parallel_bigint(str, 4, 5, 2) will return

   a[0] = 00000 = 0x0
   a[1] = 10000 = 0x10
   a[2] = 01100 = 0xC
   a[3] = 01010 = 0xA

   so a[][0] corespond to vector 1, a[][1] correspond to vector 2, etc..

   Example:

   ABCD
   7321
   09AB
   CCDE
   
   ascii_to_parallel_bigint(str, 4, 4, 16) will return

   a[0] = 0x9
   a[1] = 0xa
   a[2] = 0x3
   a[3] = 0x2
   a[4] = 0xd
   a[5] = 0x8
   a[6] = 0x3
   a[7] = 0x7
   a[8] = 0xd
   a[9] = 0x9
   a[10] = 0x6
   a[12] = 0x8
   a[12] = 0xd
   a[13] = 0x9 
   a[14] = 0xc
   a[15] = 0x7

   Note: you should ensure that number of bits in a vector don't exceed MAXBITINDEX (i.e. 512)
*/
BIGINT **ascii_to_parallel_bigint(char *str, ELEMENT length, ELEMENT num, ELEMENT base)
{
  ELEMENT ch;
  INDEX i, j, k, l, n, vsize = length;
  BIGINT **a;

  CHECK_CONDITION(num > (MAXBITINDEX+1), NULL);
  CHECK_CONDITION(base != BINARY && base != HEXADECIMAL, NULL);

  if (base == HEXADECIMAL)
    vsize *= HEXBITS;

  a = ALLOC(BIGINT *,  vsize);
  for (i = 0; i < vsize; i++)
    a[i] = bigint_new();
  
  /* binary */
  if (base == BINARY) {
    for (j = 0; j < length; j++) /* for each input */
      for (i = 0; i < num; i++)  { /* for total number of vectors */
	ch = (str[length*i+j] == '0') ? 0 : 1;
	bigint_set_bit(a[j], i, ch);
      }
  }
  /* hexadecimal */
  else {
    for (i = 0; i < num; i++) { /* for each vector */
      l = i * length;
      for (j = l; j < l + length; j++) {
	if ((ch = ascii_to_hex(str[j])) != BIGINT_ERROR)
	  for (k = 3, n = 0; k >= 0; k--, n++) {
	    bigint_set_bit(a[HEXBITS*(j%HEXBITS)+n], i, ((ch >> k) & BIT_ZERO));
	  }
	else 
	  return NIL(BIGINT *);
      }
    }
  }
  
  return a;
}

/* sets particular bits in a word of a bigint, returns BIGINT_ERROR on error, BIGINT_OK on success
   assumes a points ot a valid bigint. Performs the shifting to get the right bits so v should
   be given as a direct value. */ 
static int bigint_setbits(BIGINT *a, INDEX i, ELEMENT v, ELEMENT mask, ELEMENT shift)
{
  CHECK_CONDITION(i > MAXWORDINDEX || i < 0, BIGINT_ERROR);
  a->data[i] |= ((v << shift) & mask);
  return BIGINT_OK;
}

/* extract certain bits in a word of a bigint. returns BIGINT_ERROR on error, or the selected bits
   on success. Performas maskign first and then shift */
static int bigint_getbits(BIGINT *a, INDEX i, ELEMENT mask, ELEMENT shift)
{
  CHECK_CONDITION(i > MAXWORDINDEX || i < 0, BIGINT_ERROR);
  return ((a->data[i] & mask) >> shift);
}

/* convert an hexadecimal ASCII character [0-9A-Fa-f] into a hexadecimal integer, 
   returns BIGINT_ERROR if error, othewise the hex value */
static ELEMENT ascii_to_hex(ELEMENT c)
{
  CHECK_CONDITION(!isxdigit(c), BIGINT_ERROR);

  if (isdigit(c))
    return (c - 0x30);
  else if (islower(c))
    return (c - 0x57);
  else
   return (c - 0x37);
}
