/*=========================================================================*
 *      Title   :       majsyn.c
 *
 *      Author  :       sis
 *      Created :       Wed Oct 22 11:25:41 2003
 *
 *      Description :   Synthesis a given boolean network into a 3-input 
 *                      majority network.
 *
 * This source code can be redistributed under the GNU Free Public Licence.
 *=========================================================================*/

/* $Id: majsyn.c,v 1.6 2006/03/30 22:28:17 pgupta Exp $ */
/* $Log: majsyn.c,v $
/* Revision 1.6  2006/03/30 22:28:17  pgupta
/* added irredundant commands
/*
/* Revision 1.5  2005/09/29 20:51:18  pgupta
/* removed /home/sis from header file for compiling
/*
/* Revision 1.4  2005/07/20 16:22:49  sis
/* add more commands, make node minimum based, consider constant nodes
/*
/* Revision 1.3  2005/07/13 16:55:23  sis
/* changes make after update of purnea
/*
/* Revision 1.2  2005/06/03 18:01:48  sis
/* updates for mals and tels
/* */
/* Revision 1.1  2004/05/26 17:53:12  sis */
/* initial check in for majority synthesis, some function is also used in thsyn */
/* */

#include <stdio.h> 
#include "sis.h"
#include "../pld/pld_int.h"
#include "../simplify/simp_int.h"
#include "th_int.h"

static int  majsyn_usage ARGS(());
static int majsyn_red_remove_usage ARGS(());
static void maj_collapse ARGS((network_t *, int));
static void collapse_w_maxfanin ARGS((node_t *, int));
static int  maj_if_all_PI ARGS((node_t *));
static void network_to_majority ARGS((network_t *));
static void node_to_majority ARGS((network_t *, node_t *));
static void maj_split ARGS((network_t *, int));
static void internal_node_to_majority ARGS((network_t *, node_t *));
static int maj_variable_freq ARGS((node_t *));
static void decomp_and_or ARGS((network_t *, node_t *));
static void decomp_and ARGS((network_t *, node_t *, int));
static void decomp_or ARGS((network_t *, node_t *, int));
static void balanced_tree ARGS((int, int, node_t *(*node_func)(), node_t **, array_t **, array_t **));
static int majsyn_find_f123 ARGS((network_t *network, node_t *n));
static void majsyn_all_patterns ARGS((void));
static void maj_network_simplify ARGS((network_t *network));
static array_t *find_t_in_pfamily ARGS((pset_family f));
static array_t *find_cover_pattern ARGS((pset_family on_f, pset_family must1_set, pset_family must0_set, int allow_num, array_t *exclude_patterns));
static pset_family majsyn_simplified_pfamily ARGS((pset_family f));
static void majsyn_choose_and_or ARGS((network_t *network, node_t *n));

static st_table *st_1, *st_2, *st_3;

/***********************************************************/
/* this function is the main function that does majority   */
/* network synthesis. This given network should be first   */
/* decomposed such that no node has more than three fanins,*/
/* this can be done by using some scripts                  */
/***********************************************************/

int com_majsyn(network_t **network, int argc, char **argv)
{
  majsyn_options_t *opt;

  util_getopt_reset();

  if ((opt = majsyn_options_init(argc, argv)) == NIL(majsyn_options_t)
      || opt->help) {
    majsyn_options_free(opt);
    return majsyn_usage();
  }


  majsyn_whole_network(network, opt);
  majsyn_options_free(opt);
  
  return TH_SUCCESS;
}


/************************************************************
 * synthesis main function with redundancy removal, calls other functions from here *
 ************************************************************/
int com_majsyn_with_red_removal(network_t **network, int argc, char **argv)
{
  majsyn_options_t *opt;

  util_getopt_reset();

  if ((opt = majsyn_options_init(argc, argv)) == NIL(majsyn_options_t)
      || opt->help) {
    majsyn_options_free(opt);
    return majsyn_red_remove_usage();
  }

  // synthesize the network
  majsyn_whole_network(network, opt);

  // perform redundancy removal
  com_redundancy_removal(network, 1, "red_removal");

  // resynthezie the irredundant network
  majsyn_whole_network(network, opt);

  majsyn_options_free(opt);
  
  return TH_SUCCESS;
}




void majsyn_whole_network(network_t **network, majsyn_options_t *opt)
{

  int size = 3;
  st_generator *gen;
  char *key;
  pset_family f;
  lsGen lgen;
  node_t *n;

  /* initial collapse the nodes in the network with fanin restriction 'm' */
  if (opt->m > 0)
    (void)maj_collapse(*network, opt->m);
  /* sweep the collapsed network to get rid of redundant nodes */
  (void)network_sweep(*network);
  /* network clean up, clean all internal nodes that do not fanout to PO finally */
  network_cleanup(*network);

  /* split the network such that every node in the network has at most three variables */
  if (opt->m >0) {
    (void) maj_split(*network, size);
    /* clean up the network after maj_split */
    network_cleanup(*network);
  }
  
  /* map each node in the network into majority gate(s) */
  (void)network_to_majority(*network);

  /* clean up the network after network_to_majority */
  (void) network_cleanup(*network);
  /* sweep the collapsed network to get rid of redundant nodes */
  (void)network_sweep(*network);
  /* make sure every node is minimum based */
  foreach_node(*network, lgen, n) {
    node_minimum_base(n);
  }

#ifdef STATES_DEBUG3
  /* rui: for SEGAMT majority logic experiments */

  FILE *fp;
  char *filename;
  
  filename = ALLOC(char, 100);
  filename = util_strsav("./experiments/ga_majsyn_results.txt");
  if ((fp = fopen(filename, "w")) == NULL) 
   fail("majsyn.c: unable to open file handle for writing."); 
  //node count optimization
  fprintf(fp, "%d\n", network_num_internal(*network));   
  fclose(fp);
  if (filename != NIL(char))
    FREE(filename);
  /* end rui */
#endif
  
  st_foreach_item(st_1, gen, &key, (char **) &f)
    set_free(f);  
  st_free_table(st_1);
  st_foreach_item(st_2, gen, &key, (char **) &f)
    set_free(f); 
  st_free_table(st_2);
  st_foreach_item(st_3, gen, &key, (char **) &f)
    set_free(f);  
  st_free_table(st_3);
}  


/****************************************************/
/* this function split the given network such that  */
/* no node has more than 'size' inputs              */
/****************************************************/

static void maj_split(network_t *network, int support)
{
  split_network(network, support);
}

static void maj_network_simplify(network_t *network)
{
  int i; 
  array_t *nodevec;
  node_t *np;
  sim_accept_t accept;
  sim_method_t method;
  sim_dctype_t dctype;
  sim_filter_t filter;
  st_table *level_table;
  extern int level_node_cmp2();
    
  dctype = SIM_DCTYPE_SUB_FANIN;
  accept = SIM_ACCEPT_FCT_LITS;
  method = SIM_METHOD_SNOCOMP;
  filter = SIM_FILTER_EXACT;
  util_getopt_reset();
  simp_debug = FALSE;
  simp_trace = FALSE;
  simp_fanin_level = 1;
  simp_fanin_fanout_level = 0;

  nodevec = network_dfs(network);
  nodevec = simp_order(nodevec);
  for(i = 0; i < array_n(nodevec); i++) {
    np = array_fetch(node_t *, nodevec, i); 
    if (np->type == INTERNAL) {
      simplify_node(np, method, dctype, accept, filter, level_table); 
    }
  }
  array_free(nodevec);
}


/*******************************************************/
/* collapse the given network according to 'm'         */
/*******************************************************/
static void maj_collapse(network_t *network, int m)
{
  int i;
  array_t *node_vec;

  node_vec = network_dfs(network);
  for (i = 0; i < array_n(node_vec); i++) {
    collapse_w_maxfanin(array_fetch(node_t *,node_vec,i),m);
  }
  
  array_free(node_vec);
} 

/*****************************************************/
/* collapse the given node according to 'limit'      */
/*****************************************************/
static void collapse_w_maxfanin(node_t *node, int limit)
{
  int num_fanin, flag, i;
  node_t *copy, *fanin, *backup;

  num_fanin = node_num_fanin(node);
  /* no collapsing to be done if fanin is 0 or greater than limit */
  if ((num_fanin == 0) || (num_fanin >= limit) || 
      (node_type(node) == PRIMARY_OUTPUT) || (node_type(node) == PRIMARY_INPUT)) 
    return;
  /* while fanin is less than limit and all fanin are NOT primary inputs */
  while (num_fanin < limit && !maj_if_all_PI(node)) {
    copy = node_dup(node);
    flag = 0;
    foreach_fanin(copy, i, fanin) {
      if (node_type(fanin) != PRIMARY_INPUT) {
	if (node_num_fanout(fanin) > 0) { 
	  backup = node_dup(node);
	  node_collapse(backup, fanin);
	  
	  if (node_num_fanin(backup) <= limit &&
	      !node_equal(backup, node_constant(0)) &&
	      !node_equal(backup, node_constant(1))) {
	    node_collapse(node, fanin);
	    flag = 1;
	  }       	
	  node_free(backup);
	}
	else
	  fail("thres_collapse: the fanin node fanouts to nowhere");
      }
      num_fanin = node_num_fanin(node);
      if (num_fanin == 0)
	fail("num fanin is zero");
      if (num_fanin > limit)
	break;
    }
    node_free(copy);
    if (!flag)
      break;
  }
}


/******************************************************************************/ 
/* This function checks if all the fanins of a given node are primary inputs. */
/* It takes parameter: node_t * which is a node in sis network                */
/* If all fanin are primary inputs, return 1, otherwise 0                     */
/******************************************************************************/
static int maj_if_all_PI(node_t *node)
{
  int i, status;
  node_t *fanin;
  
  status = 1;

  foreach_fanin(node, i, fanin) {
    if (node_type(fanin) != PRIMARY_INPUT) {
      status = 0;
      break;
    }
  }
  return status;
}

/**********************************************************/
/* map the given network into majority gate based network */
/**********************************************************/
static void network_to_majority(network_t *network)
{
  int i;
  array_t *node_vec;
  node_t *n;

  /* first generate all the psets (admissible patterns) of majority functions */
  majsyn_all_patterns();
  node_vec = network_dfs(network);
  for (i = 0; i < array_n(node_vec); i++) {
    n = array_fetch(node_t *, node_vec,i);
    /* do majority synthesis only on the internal nodes */
    if (node_type(n) == INTERNAL) {
      node_to_majority(network, n);
    }
  }
   array_free(node_vec);
}

/************************************/
/* generate all admissible patterns */
/************************************/
static void majsyn_all_patterns(void)
{

  /* create the hash tables which contains all admissible patterns */
  st_1 = maj_table_1();
  st_2 = maj_table_2();
  st_3 = maj_table_3();

}

/********************************************/
/* map the given node n into majority gates */
/********************************************/
static void node_to_majority(network_t *network, node_t *n)
{
 
  if ((node_type(n) == PRIMARY_INPUT) ||(node_type(n) == PRIMARY_OUTPUT)) { 
    /* if it's a PI or PO, needn't majority mapping */
    ;
  }
  else{ /* it's an internal node, map into majority gate(s) in place */
    /* check if the node is constant 0 or 1, if yes, return without processing */
    if (node_equal(n, node_constant(0)) || node_equal(n, node_constant(1))) {
      //      printf("meet constant node: \n");
      //      node_print(sisout, n);
      return;
    }
    /**********************************************************
     * Process special case of node, such as                  *
     * 1.  f num of inputs is not 3, caused by simplification *
     * 2.  f = constant 1 or 0                                *
     *********************************************************/
    if (node_num_fanin(n) != 3)
      {
	/* do AND/OR or OR/AND decomposition on the node */
	majsyn_choose_and_or(network, n);
      }
    else{
      switch(node_function(n)) {
      case NODE_0: 
	/* itself is a majority node, do nothing, leave it in the network */
	break;
      case NODE_1: 
	/* itself is a majority node, do nothing, leave it in the network */
	break;
      case NODE_UNDEFINED:
	fail("internal_node_to_majority: node undefined");
      default: internal_node_to_majority(network, n); break;
      }    
    }
  }
}

/**************************************************/
/* given a node, this function decides whether    */
/* do AND/OR or OR/AND decompositon on this node, */
/* and decomposes this node                       */
/**************************************************/
static void majsyn_choose_and_or(network_t *network, node_t *n)
{
 int num_and_or, num_or_and;

 num_and_or = majsyn_and_or_try(n);
 num_or_and = majsyn_or_and_try(n);
 if (num_and_or <= num_or_and)
   majsyn_and_or_decomp(network, n, /* and_limit */ 2, /* or_limit */ 2);
 else
   majsyn_or_and_decomp(network, n, /* and_limit */ 2, /* or_limit */ 2);
 
}

/*******************************************************/
/* map internal node in the network into majority node */
/*******************************************************/
static void internal_node_to_majority(network_t *network, node_t *n)
{
  int variable_freq;
  int f123_exist;
  node_t *p1, *p2, *p2_lit, *fanin, *g;


  variable_freq = maj_variable_freq(n); 
  if (variable_freq == -1) { /* if every variable appears once */
    /* do AND/OR or OR/AND decomposition on the node */
    majsyn_choose_and_or(network, n);
  }
  else if (variable_freq == -2) /* if a literal doesn't exist in every product term */
    {
      /* check if the node is already a majority node */
      /* if 'yes', do nothing on this node, go on to synthesis the next node  */ 
      /* if 'no', check 'if 4 AND/OR' gate are enough for implementing this node  */
      /*          if 'yes', use AND/OR mapping and go to synthesis the next node  */ 
      /*          if 'no', use our majority synthesis method to find f123         */
      
      if (check_is_majority(n->F, st_1, st_2, st_3) == 0) { /* the node itself is not a majority gate */
	/* if <= 4 e-input AND/OR gates can implement this node, do AND/OR mapping */
	if ((majsyn_and_or_try(n) <= 4) || (majsyn_or_and_try(n) <= 4)) {
	  /* do AND/OR or OR/AND decomposition on the node */
	  majsyn_choose_and_or(network, n);
	}
	/* otherwise, do majority mapping */
	else {
	  f123_exist = majsyn_find_f123(network, n);
	  if (f123_exist !=1) { /* can't find f1, f2, f3 by our synthesis method */
	    /* do AND/OR or OR/AND decomposition on the node */
	    majsyn_choose_and_or(network, n);
	  }
	  else
	    ;
	}
      }
      else { /* the node itself is a majority node */
	;
      }
    }
  else /* a literal exists in every product term */
    {
      /* factor out this literal */
      /* e.g. f = x1' x2 + x1' x3, factor out to get f = x1' (x2 + x3) = p1 p2 */
      /* where p1 = x1' , p2 = x2 + x3 */
      
      /* now, comput p1 and p2 */
      fanin = node_get_fanin(n, variable_freq);
      switch(node_input_phase(n, fanin)) {
      case POS_UNATE:
	p1 = node_literal(fanin, 1); /* uncomplemented */
	break;
      case NEG_UNATE:
	p1 = node_literal(fanin, 0); /* complemented */
	break;
      default:
	fail("not a unate function");
      }
      p2 = node_cofactor(n, p1);
      p2_lit = node_literal(p2, 1); /* uncomplemented */
      /* add p2 into the network */
      network_add_node(network, p2);
      /* compute node n in new logic function format */
      g = node_and(p1, p2_lit);
      node_replace(n, g);
      /* newly generated node f = p1 p2 is already a majority node (2-input AND) */
      /* now we need to do majority synthesis on newly generated node p2 */
      node_to_majority(network, p2);
    }
}




/*******************************************************/
/* return: 1: Can find f1, f2, f3, and f = M(f1,f2,f3) */
/*         0: Cannot find f1, f2, f3 by our method     */
/*******************************************************/
static int majsyn_find_f123(network_t *network, node_t *n)
{
  array_t *t_minterms;
  pset_family  f1, f2, f3, must0_set, must1_set;
  pset_family must0_set_old, must1_set_old;
  array_t *exclude, *result;
  array_t *exclude_pfamilys;
  
  /************************/
  /* First step: find f1  */
  /************************/

  exclude = array_alloc(array_t *, 0);
  exclude_pfamilys = array_alloc(pset_family, 0);

  /* first check if there exists a 'T' shape with 0 or 1 making-up on-set minterm */ 
  t_minterms = find_t_in_pfamily(n->F);
  if (array_n(t_minterms) == 0) { /* couldn't find T-shape */
    /* call function 'find_cover_pattern' to find f1, */
    /* with must0_set, must1_set and exclude empty, allow_num = 7 */
    must0_set = NULL;
    must1_set = NULL;
    result = find_cover_pattern(n->F, must1_set, must0_set, 7, exclude);
    if (array_n(result) == 0)
      fail("majsyn_find_f123: can't find candidate f1");
    f1 = minterm_indexes_to_pfamily(result);
  }
  else { /* can find a T-shape */
    f1 = minterm_indexes_to_pfamily(t_minterms);
  }

  /*************************************/
  /* Second step: find f2 (first time) */
  /*************************************/
  /* compute the must0_set_old and must1_set_old based on n->F and f1 got earlier */
  f2 = NULL;
  must0_set_old = find_must_set(n->F, f1, f2, 0);
  must1_set_old = find_must_set(n->F, f1, f2, 1);
  /* find f2 */
  result = find_cover_pattern(n->F, must1_set_old, must0_set_old, 7 /* allow_num */, exclude);
  if (array_n(result) == 0){
    return 0; /* can't find f2, use AND/OR mapping */
  }
  /* already find f2 (first time) */
  f2 = minterm_indexes_to_pfamily(result);
  /*************************************/
  /* Third step: find f3 (first time)  */
  /*************************************/
  /* compute must0_set and must1_set based on n->F, f1 and f2 got earlier */
  must0_set = find_must_set(n->F, f1, f2, 0);
  must1_set = find_must_set(n->F, f1, f2, 1);
  /* find f3 */
  result = find_cover_pattern(n->F, must1_set, must0_set, 7 /* allow_num */, exclude);
  if (array_n(result) == 0){ /* first time, can't find f3 based on n->F, f1 and f2 */
    /**************************************/
    /* Second step: find f2 (second time) */
    /**************************************/
    /* backtrack, find other f2 */
    /* update the exclude array */
    array_insert_last(pset_family, exclude_pfamilys, f2);
    exclude = transform_exclude_pfamilys(exclude_pfamilys);
    result = find_cover_pattern(n->F, must1_set_old, must0_set_old, 7, exclude);
    if (array_n(result) == 0)
      return 0; /* can't find f2, use AND/OR mapping */    
    /* already find f2 (second time) */
    f2 = minterm_indexes_to_pfamily(result);
    
    /**************************************/
    /* Third step: find f3 (second time)  */
    /**************************************/    
    /* update must0_set and must1_set based on n->F, f1 and f2 got earlier */
    must0_set = find_must_set(n->F, f1, f2, 0);
    must1_set = find_must_set(n->F, f1, f2, 1);
    /* find f3 (second time) */
    result = find_cover_pattern(n->F, must1_set, must0_set, 7 /* allow_num */, exclude);
    if (array_n(result) == 0){ /* second time, can't find f3 based on n->F, f1 and f2 */
      /**************************************/
      /* Second step: find f2 (third time) */
      /*************************************/      
      /* backtrack, find other f2 */      
      /* update the exclude array */
      array_insert_last(pset_family, exclude_pfamilys, f2);
      exclude = transform_exclude_pfamilys(exclude_pfamilys);
      result = find_cover_pattern(n->F, must1_set_old, must0_set_old, 7, exclude);
      if (array_n(result) == 0) 
	return 0; /* can't find f2, use AND/OR mapping */  
      /* already find f2 (third time) */
      f2 = minterm_indexes_to_pfamily(result);
      /* update must0_set and must1_set based on n->F, f1 and f2 got earlier */
      must0_set = find_must_set(n->F, f1, f2, 0);
      must1_set = find_must_set(n->F, f1, f2, 1);
      /* find f3 (third time) */
      result = find_cover_pattern(n->F, must1_set, must0_set, 7 /* allow_num */, exclude);
      if (array_n(result) == 0) /* can't find f3 based on n->F, f1 and f2 */
	return 0; /* can't find f3, use AND/OR mapping */   /* already find f3 (third time) */
      f3 = minterm_indexes_to_pfamily(result);
    }  
    else { /* already find f3 (second time) */
      f3 = minterm_indexes_to_pfamily(result);
    }
  }
  else { /* already find f3 (first time) */
    f3 = minterm_indexes_to_pfamily(result);
  }

  /* simplify f1, f2, f3 */
  f1 = majsyn_simplified_pfamily (f1);
  f2 = majsyn_simplified_pfamily (f2);
  f3 = majsyn_simplified_pfamily (f3);
  majsyn_replace_majnode (network, n, &f1, &f2, &f3); 


  return 1;  /* successfully find f1, f2, f3 */
}

/*****************************************************************/
/* check the node to see the variable frequency                  */
/* return: -1: every literal(variable) appears exactly once      */
/*          x: a literal exists in every product term            */
/*            ( number of product term > 1 ), x is               */
/*            the index of that variable( not literal)           */
/*         -2: otherwise                                         */
/* note: pay attention to 'literal' and 'variable' here          */
/*****************************************************************/

static int maj_variable_freq(node_t *n)
{ 
  int i, result, num_fanin, max_index, max_freq, fanin_index;
  int *literals;

  /* default value */
  result = -1;
  /* get the literal count of each variable in node */
  literals = node_literal_count(n);
  /* initialize the max index to -1 and max freq to 0 */
  max_index = fanin_index = 0;
  max_freq = literals[0];

  num_fanin = node_num_fanin(n);
  for (i = 1; i < 2 * num_fanin; i++) {
    if (!(i%2))
      fanin_index++;    
    if (literals[i] > max_freq) {
      max_freq = literals[i];
      max_index = fanin_index;
    } 
  }

  if (max_freq > 1) {
    if (max_freq == node_num_cube(n))
      result = max_index;
    else 
      result = -2;
  }
  return result;
}

/****************************************************************
 * decompose a node n in the network into first level: or_limit *
 * OR and and-limit AND gates only, second level, one inverter  *
 *  only, new nodes replace the old node in the network         *
 ****************************************************************/

void majsyn_or_and_decomp(network_t *network, node_t *node, int and_limit, int or_limit)
{
  node_t *node_i, *p;
  char *name, *name_p;


  name_p = util_strsav(node_name(node));

  node_i = node_dup(node);
  node_invert(node_i);
  name = util_strsav(node_name(node_i));

  p = node_literal(node_i, 0); /* complemented */
  p->name = name_p;
  
  node_replace(node, p);
  network_add_node(network, node_i);
  majsyn_and_or_decomp(network, node_i, /* and_limit */ 2, /* or_limit */ 2);

}


/*****************************************************
 * decompose a node n in the network into and_limit  *
 * AND and or-limit OR gates only, new nodes replace *
 * the old node in the network                       *
 *****************************************************/
void majsyn_and_or_decomp(network_t *network, node_t *node, int and_limit, int or_limit)
{
  node_t *f, *tmp_c, *c;
  int i;
  
  /* needn't decomposition for PI, PO, constant 0, */ 
  /* constant 1, BUF, INV, UNDEFINED               */
  if ((node_function(node) != NODE_COMPLEX) && 
      (node_function(node) != NODE_AND) &&
      (node_function(node) != NODE_OR)) {
    return;
  }
  
  if (node->type == INTERNAL) {
    /* Check for a special case: f = a + a' will cause tech_decomp to */
    /* fail unless it is replaced by a constant. */
    if (node_function(node) == NODE_OR) {
      foreach_fanin(node, i, f) {
	if (node_input_phase(node, f) == BINATE) {
	  node_replace(node, node_constant(1));
	  break;
	}
      }
    }
    /* map this node to AND or OR gates only, no fanin restriction now */
  if (node_function(node) == NODE_AND) {
      if (node_num_fanin(node) > and_limit)
	decomp_and(network, node, and_limit);
      return;
  }
  if (node_function(node) == NODE_OR) {
    if (node_num_fanin(node) > or_limit)
      decomp_or(network, node, or_limit);
    return;
  }
  
  if (node_function(node) == NODE_COMPLEX) {
    decomp_and_or(network, node);
    /* map the above got nodes to AND or OR gates, with fanin restriction */  
    if (node_function(node) == NODE_OR && or_limit <= 0) {
      (void) node_invert(node);
    } else if (node_function(node) == NODE_AND && and_limit <= 0) {
      (void) node_invert(node);
    }
    /* map the product terms first */
    /* e.g. function f is mapped to f = f1 + f2 + f3 + f4, 
       the following code work on f1, f2, f3 and f4 respectively */

    for (i = node_num_cube(node)-1; i>= 0; i--) {
      tmp_c = dec_node_cube(node, i);
      /* pay attention here, tmp_c is got like tmp_c = f1, # of fanin is always 1 */
      /* so, in order to get correct f1, should get the 0th fanin of tmp_c */
      /* so that c = f1(x1, x2, x3) */
      c = node_get_fanin(tmp_c,0); 
      if (node_num_fanin(c) > and_limit) {
	decomp_and(network, c, and_limit);
      }
    }
    /* map the sum term second */
    switch (node_function(node)) {
    case NODE_AND:
      if (node_num_fanin(node) > and_limit)
	decomp_and(network, node, and_limit);
      break;
    case NODE_OR:
      if (node_num_fanin(node) > or_limit)
	decomp_or(network, node, or_limit);
      break;
    default:
      ;
    }
  }
  }
}


/************************************************************/
/* decomp_and_or, decomp_and, decomp_or, balanced_tree      */
/* copied from sis-1.2/sis/decomp/dec_tech.c                */
/************************************************************/
static void
decomp_and_or(network, f)
network_t *network;
node_t *f;
{
    node_t *g, *t, *clit, *c;
    int i;

    
    if (node_function(f) != NODE_COMPLEX) {
        return;
    }

    g = node_constant(0);
    for (i = node_num_cube(f)-1; i>= 0; i--) {
        c = dec_node_cube(f, i);
        network_add_node(network, c);
        clit = node_literal(c, 1);
        t = node_or(g, clit);
        node_free(clit);
        node_free(g);
        g = t;
    }
    node_replace(f, g);
}

static void decomp_and(network_t *network, node_t *f, int and_limit)
{
    array_t *nodes, *leaves;
    node_t *root, *leaf, *fanin, *node;
    int i;


    node_t *(*fp_node_and)() = node_and;
    if (and_limit <= 0) {
        fail("Error: wrong fanin limit for AND gate");
    }

    if (node_function(f) != NODE_AND) {
      fail("Error: function type is not AND in decomp_and");
    }
    if (node_num_fanin(f) <= and_limit)
      return;

    balanced_tree(node_num_fanin(f), and_limit, fp_node_and, &root, &nodes, &leaves);
 
    /* connect the leaves */
    for (i = array_n(leaves)-1; i >= 0; i--) {
        leaf = array_fetch(node_t *, leaves, i);
        fanin = node_get_fanin(f, i);
        switch (node_input_phase(f, node_get_fanin(f, i))) {
        case POS_UNATE:
            node_replace(leaf, node_literal(fanin, 1));
            break;
        case NEG_UNATE:
            node_replace(leaf, node_literal(fanin, 0));
            break;
        default:
            fail("Error: wrong phase of a input in decomp_and");
        }
    }
  /* replace f */
    for (i = array_n(nodes)-1; i >= 0; i--) {
        node = array_fetch(node_t *, nodes, i);
        if (node == root) {
            node_replace(f, node);
        } else {
            network_add_node(network, node);
        }
    }

    array_free(nodes);
    array_free(leaves);
}

static void
decomp_or(network, f, or_limit)
network_t *network;
node_t *f;
int or_limit;
{
    array_t *nodes, *leaves;
    node_t *root, *leaf, *fanin, *node;
    int i;


    node_t *(*fp_node_or)() = node_or;
    if (or_limit <= 0) {
        fail("Error: wrong fanin limit for OR gate");
    }

    if (node_function(f) != NODE_OR) {
        fail("Error: function type is not OR in decomp_or");
    }

    if(node_num_fanin(f) <= or_limit)
      return;

    balanced_tree(node_num_fanin(f), or_limit, fp_node_or, &root, &nodes, &leaves);
    
    /* connect the leaves */
    for (i = array_n(leaves)-1; i >= 0; i--) {
        leaf = array_fetch(node_t *, leaves, i);
        fanin = node_get_fanin(f, i);
        switch (node_input_phase(f, node_get_fanin(f, i))) {
        case POS_UNATE:
            node_replace(leaf, node_literal(fanin, 1));
            break;
        case NEG_UNATE:
            node_replace(leaf, node_literal(fanin, 0));
            break;
        default:
            fail("Error: wrong phase of a input in decomp_or");
        }
    }

    /* replace f */
    for (i = array_n(nodes)-1; i >= 0; i--) {
      node = array_fetch(node_t *, nodes, i);
      if (node == root) {
	node_replace(f, node);
      } else {
	network_add_node(network, node);
      }
    }
    array_free(nodes);
    array_free(leaves);
}

static void balanced_tree(n, m, node_func, root, nodes, leaves)
int n;                          /* number of leaves */
int m;                          /* branching factor */
node_t *(*node_func)();         /* function at a node */
node_t **root;                  /* points to the root of the tree */
array_t **nodes;                /* points to all nodes */
array_t **leaves;               /* points to the leaves */
{
    node_t *np, *rt;
    array_t *ns, *lv;
    int d, s;
    bool first;

    if (n == 1) {
        *root = node_alloc();
        *nodes = array_alloc(node_t *, 0);
        array_insert_last(node_t *, *nodes, *root);
        *leaves = array_alloc(node_t *, 0);
        array_insert_last(node_t *, *leaves, *root);
    } else {
        first = TRUE;
        d = (n + m - 1) / m;
        while (n > 0) {
            s = MIN(d, n);
            n = n - d;
           balanced_tree(s, m, node_func, &rt, &ns, &lv);
            if (first) {
                *root = node_literal(rt, 1);
                *nodes = array_alloc(node_t *, 0);
                array_insert_last(node_t *, *nodes, *root);
                array_append((*nodes), ns);
                array_free(ns);
                *leaves = lv;
                first = FALSE;
            } else {
                np = node_literal(rt, 1);
                node_replace(*root, (*node_func)(*root, np));
                node_free(np);
                array_append((*nodes), ns);
                array_append((*leaves), lv);
               array_free(ns);
                array_free(lv);
            }
        }
    }
}


/* check if a node can be decoposed into <= 4 AND/OR(2-input) nodes */
int majsyn_and_or_try(node_t *n)
{ 
  int node_num = 0, i;
  node_t *tmp_node;
  network_t *tmp_network;
  array_t *node_array;
  node_t *node;
  
  tmp_node = node_dup(n);
  tmp_network = network_create_from_node(tmp_node);

  /* do AND/OR decomposition on the node */
  node_array = network_dfs(tmp_network);
  if (network_num_internal(tmp_network) == 1) {
    for (i = 0; i < array_n(node_array); i++) {
      node = array_fetch(node_t *, node_array, i);
      if ((node_type(node) != PRIMARY_OUTPUT) && (node_type(node) != PRIMARY_INPUT)) { 
	majsyn_and_or_decomp(tmp_network, node, /* and_limit */ 2, /* or_limit */ 2);
	network_sweep(tmp_network);
	node_num =  network_num_internal(tmp_network);
	break;
      }
    }
  }   
  else 
    fail("majsyn_and_or_try: node number is incorrect");

  network_free(tmp_network);

  return node_num;
}
  

/* check if a node can be decoposed into <= 4 OR/AND(2-input) nodes */
int majsyn_or_and_try(node_t *n)
{ 
  int node_num = 0, i;
  node_t *tmp_node;
  network_t *tmp_network;
  array_t *node_array;
  node_t *node;
  
  tmp_node = node_dup(n);
  /* invert the node first, then do AND/OR mapping on the inverted node */

  node_invert(tmp_node);
  tmp_network = network_create_from_node(tmp_node);

  /* do AND/OR decomposition on the node */
  node_array = network_dfs(tmp_network);
  if (network_num_internal(tmp_network) == 1) {
    for (i = 0; i < array_n(node_array); i++) {
      node = array_fetch(node_t *, node_array, i);
      if ((node_type(node) != PRIMARY_OUTPUT) && (node_type(node) != PRIMARY_INPUT)) { 
	majsyn_and_or_decomp(tmp_network, node, /* and_limit */ 2, /* or_limit */ 2);
	network_sweep(tmp_network);
	node_num =  network_num_internal(tmp_network);
	break;
      }
    }
  }   
  else 
    fail("majsyn_and_or_try: node number is incorrect");

  network_free(tmp_network);
  node_num = node_num + 1; /* also count the first inverter */

  return node_num;
}
  
/************************************************/
/* Given a pset_family f, check if there exists */
/* some 'T' shape in it if allowing zero or one */
/* make-up on set minterm                       */
/************************************************/
static array_t *find_t_in_pfamily(pset_family f)
{
  int count, i;
  int makeup;
  char *key;
  array_t *t_minterms, *f_minterms, *t_minterms_makeup;
  st_generator *gen;
  pset_family f_in_st;
  
  makeup = 0;  /* initialize as no makeup */
  /* t_indexes saves the minterm indexes of the found 'T' shape */
  t_minterms = array_alloc(int,0);
  /* get minterm indexes of f in an array */
  f_minterms = pfamily_minterms(f);  
  /* in 'T'-shape hash table, find desired 'T' */
  st_foreach_item(st_3, gen, &key, (char **) &f_in_st) {
    count = 0;
    t_minterms = pfamily_minterms(f_in_st);
    /* check if this T satisfies */
    for (i = 0; i < array_n(f_minterms); i++) {
      /* if this minterm is a minterm in this T */
      if (majsyn_array_find_element(t_minterms, array_fetch(int, f_minterms, i)))
	count++;
      if (count == 4) 
	break;
      if ((count == 3) && (makeup == 0)) {
	makeup = 1; /* there is one makeup onset-minterm */
	/* save the makeup T minterms and continue finding better T*/
	t_minterms_makeup = array_dup(t_minterms);
      }
    }
    if (count == 4)
      break;
    count = 0;
  }

  if (count == 4) {
    if (makeup == 1)
      array_free(t_minterms_makeup);
    return t_minterms;
  }
  else if (makeup == 1)
    return t_minterms_makeup;
  
else {
    t_minterms = array_alloc(int, 0);
    return t_minterms;
  }
}

/***************************************************/
/* Given a pset_family f for the original funciton */
/* and must1_set, must0_set and exclude_patterns   */
/* return an admissible-pattern pset_family, which */
/* have all minterms in must1_set, and cannot have */
/* any minterm in must0_set, and cannot be in the  */
/* exclude_patterns. There maybe exists many such  */
/* pset_familys, choose the one with the least     */
/* number of makeup one (i.e. not in both on_f and */
/* must1_set). There may still exists more then one*/
/* solution, choose the one that found earlier.    */
/* return an array which contains the indexes of   */
/* the minterms covered by the pset_family         */
/***************************************************/
static array_t *find_cover_pattern(pset_family on_f,
				   pset_family must1_set, 
				   pset_family must0_set, 
				   int allow_num,
				   array_t *exclude_patterns)
{
  int i, j, flag, makeup_num, least_makeup;
  int m1, m0, ma;
  array_t *on_minterms, *must1_minterms, *must0_minterms;
  array_t *all_patterns;  
  array_t *admissible_indexes, *candidate_indexes;
  char *key;
  st_generator *gen;
  pset_family f;
  
  /* initialize least_makeup */
  least_makeup = 10;
  makeup_num = 10;
  on_minterms = pfamily_minterms(on_f);
  if (must1_set != NULL) 
    must1_minterms = pfamily_minterms(must1_set); 
  else
    must1_minterms = array_alloc(int, 0);
  if (must0_set != NULL)
    must0_minterms = pfamily_minterms(must0_set); 
  else
    must0_minterms = array_alloc(int, 0);

  /* get all pset_family of admissible patterns in array: all_patterns */
  all_patterns = array_alloc(pset_family, 0);
  st_foreach_item(st_3, gen, &key, (char **) &f) {
    array_insert_last(pset_family, all_patterns, f);
  }
  st_foreach_item(st_2, gen, &key, (char **) &f) {
    array_insert_last(pset_family, all_patterns, f);
  }
  st_foreach_item(st_1, gen, &key, (char **) &f) {
    array_insert_last(pset_family, all_patterns, f);
  }

/* for each admissible pattern (saved in all_patterns array),   */
/* check if it covers must1_minterms, and not in must0_minterms */
  for (i = 0; i < array_n(all_patterns); i++) {
    flag = 1; /* initialize as 'find the cover' */
    f = array_fetch(pset_family, all_patterns, i); /* get the admissible pattern */
    admissible_indexes = pfamily_minterms(f); /* get the indexes of this admissible pattern */
    if ( must1_set != NULL) { /* if must1_set is non-empty, go on check */
      for (j = 0; j < array_n(must1_minterms); j++) { /* check every minterm in must1_set */
	m1 = array_fetch(int, must1_minterms, j);
	if (majsyn_array_find_element(admissible_indexes, m1) == 0) {
	  flag = 0;
	  break; /* jump out of checking must1_set */
	}
      }
    }
    
    if (flag == 1) { /* check if this admissible pattern violates must0_set */
      if ( must0_set != NULL) { /* if must0_set is non-empty, go on check */
 	for (j = 0; j < array_n(must0_minterms); j++) {
	  m0 = array_fetch(int, must0_minterms, j);
	  if (majsyn_array_find_element(admissible_indexes, m0) == 1) {
	    flag = 0;
	    break; /* violate must0_set, jump out of checking must0_set */
	  }
	}
      }
    }
    if (flag == 1) { /* count the makeup_num */
      makeup_num = 0;
      /* for each minterm in admissible_indexes: ma, check if it's a makeup 1 */
      for (j = 0; j < array_n(admissible_indexes); j++) { 
	ma = array_fetch(int, admissible_indexes, j);
	if ((majsyn_array_find_element(on_minterms, ma) == 0)
	    && (majsyn_array_find_element(must1_minterms, ma) == 0))
	  makeup_num++; /* increase the makeup_num by one */
      }
      if (makeup_num < least_makeup) { /* take this admissible pattern as candidate */
	  if ((array_n(exclude_patterns) == 0) || (pattern_is_element(exclude_patterns, admissible_indexes) == 0)) {
	    least_makeup = makeup_num; /* update least_makeup */
	    candidate_indexes = array_dup(admissible_indexes);
	  }
      }
    }
  }

  if ((least_makeup >= 8) || (least_makeup > allow_num))
    candidate_indexes = array_alloc(int, 0);

  return candidate_indexes; 
}


/************************************************/
/* compute the simplified form of a pset_family */
/* of admissible patterns and returns the       */
/* simplified pset_family                       */
/************************************************/
static pset_family majsyn_simplified_pfamily (pset_family f)
{ 
  char *key;
  array_t *minterms;
  pset_family pattern;


  /* get the unique, sorted minterm indexes */
  minterms = pfamily_minterms(f);

  /* transform the indexes into key in hashtable */
  key = majsyn_create_key(minterms);
  /* search in hash table */
  if (st_lookup(st_1, key, (char **) &pattern))
    {
      FREE(key);
      return pattern;
    }
  if (st_lookup(st_2, key, (char **) &pattern))
    {
      FREE(key);
    return pattern;
    }
  if (st_lookup(st_3, key, (char **) &pattern))
    {
      FREE(key);
    return pattern;
    }
  fail("majsyn_simplified_pfamily: This pset_family is not an admissible  pattern");

}



/***************************************/
/* usage information of majsyn command */
/***************************************/
static int majsyn_usage(void)
{
  fprintf(siserr, "usage: majsyn [-n int1]\n");
  fprintf(siserr, "    -n int1 \tInitial collapse fanin restriction set to int1 (default no collapsing).\n");
  return TH_FAIL;
}

/***************************************/
/* usage information of majsyn_red_remove command */
/***************************************/
static int majsyn_red_remove_usage(void)
{
  fprintf(siserr, "usage: majsyn_red_remove [-n int1]\n");
  fprintf(siserr, "    -n int1 \tInitial collapse fanin restriction set to int1 (default no collapsing).\n");
  return TH_FAIL;
}
