/*=========================================================================*
 *	File    : thatpg.c
 *
 *	Author  : Pallav Gupta
 *	Created : Sat Dec 27 00:34:53 2003
 *
 * This source code can be redistributed under the GNU Free Public Licence.
 *=========================================================================*/

/* $Id: thatpg.c,v 1.4 2004/01/22 02:21:11 pgupta Exp $ */

#include "sis.h"
#include "th_int.h"

EXTERN char *thatpg_create_vector ARGS((network_t *, int));

static int thatpg_usage ARGS(());

/* threshold automatic test pattern generation */
int com_thatpg(network_t **network, int argc, char **argv)
{
  thatpg_info_t *info;
  thatpg_options_t *opt;
  lsGen gen;
  node_t *n;

  /* check if threshold network */
  if (is_threshold_network(*network) != TH_SUCCESS)
    return TH_FAIL;

  /* check if ampl/cplex are present */
  if (util_path_search(AMPL) == NIL(char)) {
    fprintf(siserr, "AMPL not found. Please download from www.ampl.com and install.\n");
    return TH_FAIL;
  }
  if (util_path_search(ILP_SOLVER) == NIL(char)) {
    fprintf(siserr, "CPLEX not found. Please download from www.cplex.com and install.\n");
    return TH_FAIL;
  }

  /* initialize data structures */
  info = thatpg_info_init(*network);
  opt = thatpg_options_init(argc, argv);
  info->thatpg_opts = opt;

  if (opt == NIL(thatpg_options_t)) {
    thatpg_info_free(info);
    return TH_FAIL;
  }
  else if (opt->help) {
    thatpg_info_free(info);
    return thatpg_usage();
  }
  /* TODO: FSM test generation */
  else if (info->seq) {
    fprintf(siserr, "We don't support FSM test generation just yet\n");
    thatpg_info_free(info);
    return TH_FAIL;
  }

  /* Generate all the faults */
  fprintf(siserr, "Generating fault list ...\n");
  if (thatpg_gen_faults(info) != TH_SUCCESS) {
    fprintf(siserr, "Error in generating faults\n");
    thatpg_info_free(info);
    return TH_FAIL;
  }
  if (info->thatpg_opts->verb > 2)
    thatpg_print_all_faults(siserr, info);

  /* calculate the propagation cubes */
  fprintf(siserr, "Generating propagation cubes ...\n");
  TH_ASSERT(thatpg_calc_pcubes(*network) == TH_SUCCESS);

  /* calculate the singular covers */
  /* fprintf(siserr, "Generating singular covers ...\n"); */
  /* TH_ASSERT(thatpg_calc_scovers(*network) == TH_SUCCESS); */

  /* create the fanin lines */
  fprintf(siserr, "Creating fanin lines ...\n");
  foreach_node(*network, gen, n)
    thnode_create_fanin_lines(n);

  /* create a test vetor table */
  info->tests = st_init_table(strcmp, st_strhash);

  /* Do random test generation first */
  fprintf(siserr, "Performing random test generation ...\n");
  thatpg_random_test_generation(info);
  
  /* Perform deterministic test generation now */
  fprintf(siserr, "Performing deterministic test generation ...\n");
  thatpg_deterministic_test_generation(info);
  thatpg_dump_statistics(info);

  /* Print the test set */
  thatpg_print_test_vectors(info);

  /* free and cleanup */
  thatpg_info_free(info);
  return TH_SUCCESS;
}


/* generate random test vectors and do fault simultion and remove detected faults from fault list*/
int thatpg_random_test_generation(thatpg_info_t *info)
{
  char *rvec, *buf;
  BIGINT **bint;
  int totalf, detectf, i;
  double fcoverage;
  st_table *faults;
  st_generator *gen;
  thfault_t *f;
  static long int time = 0;
 
   /* number of total/detected faults, fault coverage */
  faults = info->fault_table;
  totalf = st_count(faults);
  detectf = 0;
  fcoverage = 0.0;

  /* seed the random number generator */
  srand(time_seed());
  time = (util_cpu_time() - time)/1000;

  /* while fault coverage is below threshold and we have faults to detect and the timeout hasn't expired*/
  while ((fcoverage < info->thatpg_opts->fcoverage_limit) && (st_count(faults) > 0) && (time < info->thatpg_opts->rand_timeout)) {
    info->simv = array_alloc(BIGINT **, 0);
    buf = ALLOC(char, info->thatpg_opts->nvec * info->nrpi + 1);
    memset(buf, 0, info->thatpg_opts->nvec * info->nrpi + 1);

    /* generate random vectors */
    for (i = 0; i < info->thatpg_opts->nvec; i++) {
      rvec = random_bitvector(info->nrpi);
      if (!st_insert(info->tests, rvec, (char *) 0))
	memcpy(&buf[i * info->nrpi], rvec, info->nrpi);
    }
    
    /* convert them into parallel bigints */
    info->nv = info->thatpg_opts->nvec;
    TH_ASSERT((bint = ascii_to_parallel_bigint(buf, info->nrpi, info->nv, BINARY)) != NIL(BIGINT *));
    array_insert_last(BIGINT **, info->simv, bint);

    /* simulate fault free circuit */
    TH_ASSERT((info->simr = thfsim_simulate_vectors(info)) != NIL(array_t));    
    /* simulate all the faults */
    st_foreach_item(faults, gen, (char **) &f, (char **) 0) {
      if (f->status == TH_ATPG_TESTED) continue;
      if (info->thatpg_opts->verb > 1)
	thatpg_print_fault(siserr, f);

      thfsim_simulate_fault(info, f);
      if (f->status == TH_ATPG_TESTED)
	detectf++;
    }

    /* free space */
    array_bigint_free(info->simv, info->nrpi);
    array_bigint_free(info->simr, info->nrpo);
    info->simv = info->simr = NIL(array_t);
    FREE(buf);

    /* update time and fault coverage*/
    time = (util_cpu_time() - time)/1000;
    fcoverage = (double) detectf/totalf;
    break;
  }

  return TH_SUCCESS;
}

/* create the ilp models that we use to find a vector to excite a fault*/
int thatpg_fault_ilpmodels(thatpg_info_t *info)
{
  FILE *f1, *f2;
  
  if ((f1 = com_open_file(info->ilpmodel1_file, "w", NIL(char *), 0)) == NIL(FILE))
    return TH_FAIL;
  
  if ((f2 = com_open_file(info->ilpmodel2_file, "w", NIL(char *), 0)) == NIL(FILE))
    return TH_FAIL;

  /* parameters */
  fprintf(f1, "set X;\n");
  fprintf(f2, "set X;\n");
  fprintf(f1, "param w {j in X};\n");
  fprintf(f2, "param w {j in X};\n");
  fprintf(f1, "param t;\n");
  fprintf(f2, "param t;\n");
  fprintf(f1, "param don;\n");
  fprintf(f2, "param don;\n");
  fprintf(f1, "param doff;\n");
  fprintf(f2, "param doff;\n");
  fprintf(f1, "param i;\n");
  fprintf(f2, "param i;\n");
  fprintf(f1, "param delta;\n");
  fprintf(f2, "param delta;\n");
  /* decision variables, i.e. assignments */
  fprintf(f1, "var x {j in X} binary;\n");
  fprintf(f2, "var x {j in X} binary;\n");
  /* summation */
  fprintf(f1, "maximize total: sum {j in X: j != i} (w[j] * x[j]);\n");
  fprintf(f2, "maximize total: sum {j in X: j != i} (w[j] * x[j]);\n");
  /* equation 6a in paper */
  fprintf(f1, "subject to const1: (t-w[i]+don) <= sum {j in X: j != i } (w[j] * x[j]) <= (t-doff-delta);\n");
  /* equation 6b in paper */
  fprintf(f2, "subject to const1: (t+don) <= sum {j in X: j != i } (w[j] * x[j]) <= (t-doff-w[i]-delta);\n");
  fclose(f1);
  fclose(f2);
  return TH_SUCCESS;
}

/* create the data to solve an ilp problem for a given fault */
int thatpg_fault_ilpdata(thatpg_info_t *info, thfault_t *f)
{
  FILE *fp;
  int i;
  thatpg_options_t *opt = info->thatpg_opts;
  node_t *n = f->node;
  thnode_t *th = (thnode_t *) (n->th);

  if ((fp = com_open_file(info->ilpdata_file, "w", NIL(char *), 0)) == NIL(FILE))
    return TH_FAIL;

  /* we dont' want to deal with any output/stem faults, because stem faults 
     should never be present if we're doing fault collapsing, if we aren't, 
     then for each branch of the stem, we test that fault */
  TH_ASSERT(f->findex != -1);

  /* the indices */
  fprintf(fp, "set X :=");
  for (i = 0; i < node_num_fanin(n); i++)
    fprintf(fp, " %d", i);
  fprintf(fp, ";\n");
  /* the weights */
  fprintf(fp, "param w :=");
  for (i = 0; i < node_num_fanin(n); i++)
    fprintf(fp, " %d %f", i, th->W[i]);
  fprintf(fp, ";\n");
  /* threshold, don/doff, delta (required to prevent errors from ilp in double inequality) */
  fprintf(fp, "param t := %f;\n", th->T);
  fprintf(fp, "param don := %f;\n", opt->don);
  fprintf(fp, "param doff := %f;\n", opt->doff);
  fprintf(fp, "param delta := 0.001;\n");
  /* which index (input) we're testing on */
  fprintf(fp, "param i := %d;\n", f->findex);

  /* initialze all the decisin variables to -1. the reason for this is, it is very difficult
   right now for ampl to return exit codes and for me to parse to figure out if we found a valid solution
   or not. So the hack is we initialize al the decision variables to -1. If a solution is found, all the variables
   will have 0/1 value. if any has -1 this means the solution was not found. 

   TODO: I know this is a lousy hack and it might create bugs, but this is the best I can think of
   until I come up with a better scheme or ampl starts to return status codes of the solver (which might
   happen in a future release of AMPL) */
  fprintf(fp, "param x :=");
  for (i = 0; i < node_num_fanin(n); i++) {
    if (i == f->findex)
      fprintf(fp, " %d %d", i, 0);
    else
      fprintf(fp, " %d %d", i, -1);
  }
  fprintf(fp, ";\n");
  fclose(fp);
  return TH_SUCCESS;
}


/* returns a filename which contains all the ilp commands to execute */
char *thatpg_ilpcommands(char *ilpmodel, char *ilpdata, char *ilpresult)
{
  char *tmpfile;
  FILE *fp;

  /* temp file */
  tmpfile = util_tempnam(NIL(char), "SIS");
  if ((fp = com_open_file(tmpfile, "w", NIL(char *), 0)) == NIL(FILE))
    return NIL(char);
  
  fprintf(fp, "model %s;\n", ilpmodel);
  fprintf(fp, "data %s;\n", ilpdata);
  fprintf(fp, "option solver_msg 0;\n");
  fprintf(fp, "option solver %s;\n", ILP_SOLVER);
  fprintf(fp, "solve >/dev/null;\n");
  fprintf(fp, "print {j in X} x[j] >%s;\n", ilpresult);

  fclose(fp);
  return tmpfile;
}

/* reads an ilp result file and returns an array of values (0/1) for all the fanin of a node */
array_t *thatpg_read_ilpresult(char *filename)
{
  array_t *a;
  char *line, *word;
  int linelen = 0, num;
  FILE *fp;

  /* open file */
  if ((fp = com_open_file(filename, "r", NIL(char *), 0)) == NIL(FILE))
    return NIL(array_t);
  
  /* read a line format is like 0 1 0 0 ... */
  a = array_alloc(int, 0);
  line = word = NIL(char);
  while(read_line(fp, &line, &linelen) != EOF) {
    /* convert each integer and put in array */
    while ((word = gettoken(line, &line)) != NIL(char)) {
      if ((num = atoi(word)) == -1) {
	array_free(a);
	fclose(fp);
	return NIL(array_t);
      }
      array_insert_last(int, a, num);
    }
  }

  fclose(fp);
  return a;
}

/* print values of the network */
void thatpg_print_values(network_t *network)
{
  lsGen gen;
  node_t *n;
  int i;

  foreach_node(network, gen, n) {
    node_print(siserr, n);
    if (node_num_fanin(n) > 1) {
      fprintf(siserr, "%s ", thatpg_convert5(TH_GETVALUE(n)));
      for (i = 0; i < node_num_fanin(n); i++)
	fprintf(siserr, "%s ", thatpg_convert5(TH_ATPG_GETLINEVALUE(n, i)));
      fprintf(siserr, "\n");
    }
    else
      fprintf(siserr, "%s: %s\n", node_name(n), thatpg_convert5(TH_GETVALUE(n)));
  }
  
  fprintf(siserr, "\n");
}

/* convert 5 value system to string */
char *thatpg_convert5(thatpg_simv_t v)
{
  static char *s[5] = {"0", "1", "D", "DBAR", "X"};
  
  switch (v) {
  case TH_ATPG_ZERO: return s[0];
  case TH_ATPG_ONE: return s[1];
  case TH_ATPG_D: return s[2];
  case TH_ATPG_DBAR: return s[3];
  case TH_ATPG_X: return s[4];
  default: fail("unrecognized value");
  }
}

/* initialize network lines */
void thatpg_init_lines(network_t *network)
{
  lsGen gen;
  node_t *n;
  int i;

  foreach_node(network, gen, n) {
    TH_SETVALUE(n, TH_ATPG_X);
    for (i = 0; i < node_num_fanin(n); i++)
      TH_ATPG_SETLINEVALUE(n, i, TH_ATPG_X);
    TH_SETSIMCHANGED(n, 0);
  }
}

/* perform deterministic test generation */
int thatpg_deterministic_test_generation(thatpg_info_t *info)
{
  char *vec, *com1file, *com2file, buf[TH_MAXWORD] = {0}, buf1[TH_MAXWORD] = {0};
  st_table *faults = info->fault_table;
  thfault_t *f, *f1;
  st_generator *gen, *gen1;
  array_t *result;
  int i;
  BIGINT **bint;

  /* create the ilp models and command files and system commands to exec*/
  TH_ASSERT(thatpg_fault_ilpmodels(info) != TH_FAIL);
  TH_ASSERT((com1file = thatpg_ilpcommands(info->ilpmodel1_file, info->ilpdata_file, info->ilpresult_file)) != NIL(char));
  TH_ASSERT((com2file = thatpg_ilpcommands(info->ilpmodel2_file, info->ilpdata_file, info->ilpresult_file)) != NIL(char));  
  sprintf(buf, "%s %s 2>/dev/null", AMPL, com1file);
  sprintf(buf1, "%s %s 2>/dev/null", AMPL, com2file);

  /* for each untested fault */
  st_foreach_item(faults, gen, (char **) &f, (char **) 0) {
    /* TODO: handle stem faults */
    if (f->status != TH_ATPG_UNTESTED || f->type == TH_ATPG_STEM) continue;

    /* create the ilp data for this fault */
    TH_ASSERT(thatpg_fault_ilpdata(info, f) != TH_FAIL);

    /* run the solver on the first file */
    system(buf);
    if ((result = thatpg_read_ilpresult(info->ilpresult_file)) == NIL(array_t)) {
      /* nope, run on second file */
      system(buf1);
      /* both failed, fault is redundant since it can't be excited */
      if ((result = thatpg_read_ilpresult(info->ilpresult_file)) == NIL(array_t)) {
	f->status = TH_ATPG_REDUNDANT;
	continue;
      }
    }

    /* print fault */
    if (info->thatpg_opts->verb > 1)
      thatpg_print_fault(siserr, f);
    
    /* print what ilp solved */
    if (info->thatpg_opts->verb > 2) {
      fprintf(siserr, "Fault excitaiton vector = ");
      for (i = 0; i < array_n(result); i++)
	fprintf(siserr, "%d ", array_fetch(int, result, i));
      fprintf(siserr, "\n");
    }

    /* initialize the network lines */
    thatpg_init_lines(info->network);

    /* excite the fault, if not possible, fault is redundant */
    thatpg_excite_fault(f, result);
        
    if (info->thatpg_opts->verb > 2) {
      fprintf(siserr, "Exciting fault ...\n");
      thatpg_print_values(info->network);
    }
    
    /* perform test generation */
    if (thatpg_podem(info, f) == TH_SUCCESS) {
      f->status = TH_ATPG_TESTED;
      
      /* vector found, add to list */
      vec = thatpg_create_vector(info->network, info->nrpi);
      st_insert(info->tests, vec, (char *) 0);

      /* perform fault simulation */
      info->simv = array_alloc(BIGINT **, 0);
      info->nv = 1;
      TH_ASSERT((bint = ascii_to_parallel_bigint(vec, info->nrpi, info->nv, BINARY)) != NIL(BIGINT *));
      array_insert_last(BIGINT **, info->simv, bint);
      /* simulate fault free circuit */
      TH_ASSERT((info->simr = thfsim_simulate_vectors(info)) != NIL(array_t));    
      /* simulate all the faults */
      st_foreach_item(faults, gen1, (char **) &f1, (char **) 0) {
	if (f1->status != TH_ATPG_UNTESTED) continue;
	thfsim_simulate_fault(info, f1);
      }
      array_bigint_free(info->simv, info->nrpi);
      array_bigint_free(info->simr, info->nrpo);
      info->simv = info->simr = NIL(array_t);
    }
    else
      f->status = TH_ATPG_ABORTED;
   
    array_free(result);
  }
  
  unlink(com1file);
  unlink(com2file);
  FREE(com1file);
  FREE(com2file);

  return TH_SUCCESS;
}

/* perform test generation */
int thatpg_podem(thatpg_info_t *info, thfault_t *f)
{
  thatpg_assignment_t *obj, *asgn;
  bool success = FALSE;

  /* if exceeded number of backtracks */
  if (f->backtracks >= info->thatpg_opts->backtrack_limit)
    return TH_FAIL;

  if (thatpg_output_error(info->network))
    return TH_SUCCESS;

  if (!thatpg_test_possible(f)) {
    f->backtracks++;
    return TH_FAIL;
  }

  obj = thatpg_objective(f);
  if (obj == NIL(thatpg_assignment_t))
    return TH_FAIL;

  asgn = thatpg_backtrace(obj);
  if (asgn == NIL(thatpg_assignment_t)) {
    FREE(obj);
    FREE(asgn);
    return TH_FAIL;
  }
  
  thatpg_imply(f, asgn, 1);
  if (thatpg_podem(info, f) == TH_SUCCESS)
    success = TRUE;
  
  /* reverse assignment */
  if (!success) {
    if (asgn->v == TH_ATPG_ONE)
      asgn->v = TH_ATPG_ZERO;
    else
      asgn->v = TH_ATPG_ONE;  
    thatpg_imply(f, asgn, 1);
    if (thatpg_podem(info, f) == TH_SUCCESS)
      success = TRUE;
  }
  
  /* conflict both ways, make it don't care */
  if (!success) {
    asgn->v = TH_ATPG_X;
    thatpg_imply(f, asgn, 1);
  }

  FREE(obj);
  FREE(asgn);

  if (success)
    return TH_SUCCESS;
  else {
    f->backtracks++;
    return TH_FAIL;
  }
}


/* excite the fault by making assignments to pi */
int thatpg_excite_fault(thfault_t *f, array_t *result)
{
  thatpg_assignment_t *p, *k;
  int i;

  /* make assignments to excite fault */
  p = ALLOC(thatpg_assignment_t, 1);
  for (i = 0; i < node_num_fanin(f->node); i++) {
    if (i == f->findex) continue; // todo, handle stem cases

    p->n = node_get_fanin(f->node, i);
    switch (array_fetch(int, result, i)) {
    case 0: p->v = TH_ATPG_ZERO; break;
    case 1: p->v = TH_ATPG_ONE; break;
    default: fail("illegal value in ilp result");
    }
    k = thatpg_backtrace(p);
    if (k != NIL(thatpg_assignment_t)) {
      thatpg_imply(f, k, 1);
      FREE(k);
    }
    else {
      FREE(p);
      return TH_FAIL;
    }
  }
  
  /* the fault itself */
  p->n = f->fanin;
  if (f->type == TH_ATPG_NORMAL) {
    p->v = (f->value == TH_ATPG_SA0) ? TH_ATPG_D : TH_ATPG_DBAR;
    k = thatpg_backtrace(p);
    if (k != NIL(thatpg_assignment_t))
      thatpg_imply(f, k, 1);
    else {
      FREE(p);
      return TH_FAIL;
    }
  }
  else if (f->type == TH_ATPG_BRANCH) {
    p->v = (f->value == TH_ATPG_SA0) ? TH_ATPG_ONE : TH_ATPG_ZERO;
    k = thatpg_backtrace(p);
    if (k != NIL(thatpg_assignment_t)) {
      thatpg_imply(f, k, 1);
      p->v = (f->value == TH_ATPG_SA0) ? TH_ATPG_D : TH_ATPG_DBAR;
      thatpg_imply(f, p, 0);
    }
    else {
      FREE(p);
      return TH_FAIL;
    }
  }
  else {
    ; // todo stem case
  }
    
  FREE(p);
  FREE(k);
  return TH_SUCCESS;
}

/* return an objective to help in finding a test pattern, if possible */
thatpg_assignment_t *thatpg_objective(thfault_t *f)
{
  pset_family pcubes;
  pset cube;
  thatpg_assignment_t *p = ALLOC(thatpg_assignment_t, 1);
  int i, idx;
  thatpg_simv_t v;
  bool d;
  node_t *n;

  TH_ASSERT(lsLength(f->dfront) > 0);
  lsDelBegin(f->dfront, (lsGeneric *) &n);

  //  fprintf(siserr, "propagating through %s\n", node_name(n));

  /* find an input on which d or dbar exists */
  idx = -1;
  for (i = 0; i < node_num_fanin(n); i++) {
    v = TH_ATPG_GETLINEVALUE(n, i);
    if (v == TH_ATPG_D || v == TH_ATPG_DBAR) {
      idx = i;
      d = (v == TH_ATPG_D) ? TRUE : FALSE;
      break;
    }
  }
  /* none found, fail */
  if (idx == -1) {
    FREE(p);
    return NIL(thatpg_assignment_t);
  }

  /* get the propagation cubes for this input */
  pcubes = ((thnode_t *) n->th)->pcubes[idx];
  /* pick a cube that will propagate the appropriate d or dbar */
  for (i = 0; i < pcubes->count; i++) {
    cube = GETSET(pcubes, i); 
    /* this cube propagates a d */
    if ((d && (GETINPUT(cube, idx) == ZERO)) ||
	/* this cube propagates a dbar */
	(!d && (GETINPUT(cube, idx) == ONE)))
      break;
  }
  
  /* pick an input with x */
  for (i = 0; i < node_num_fanin(n); i++)
    if (TH_GETVALUE(node_get_fanin(n, i)) == TH_ATPG_X)
      break;
  /* no such input exists fail */
  if (i == node_num_fanin(n)) {
    FREE(p);
    return NIL(thatpg_assignment_t);
  }

  /* get the bit from cube for assignment */
  if (GETINPUT(cube, i) == ZERO)
    v = TH_ATPG_ZERO;
  else
    v = TH_ATPG_ONE;
  
  p->n = node_get_fanin(n, i);
  p->v = v;

  //  fprintf(siserr, "objective: %s %s\n", node_name(p->n), (p->v == TH_ATPG_ZERO) ? "0" : "1");
  return p;
}


/* map an objective into a primary assignment if possible. */
thatpg_assignment_t *thatpg_backtrace(thatpg_assignment_t *p)
{
  thatpg_assignment_t *k = ALLOC(thatpg_assignment_t, 1);
  thatpg_simv_t v;
  int i;

  k->n = p->n;
  k->v = p->v;
  
  while (node_type(k->n) != PRIMARY_INPUT) {
    /* inversion of output */
    switch (k->v) {
    case TH_ATPG_ONE: v = TH_ATPG_ZERO; break;
    case TH_ATPG_ZERO: v = TH_ATPG_ONE; break;
    case TH_ATPG_D: v = TH_ATPG_ZERO; break;
    case TH_ATPG_DBAR: v = TH_ATPG_ONE; break;
    default:
      fail("illegal value in podem backtrace");
    }
    /* compute k->v = k->v XOR v */
    if ((v == TH_ATPG_ONE && k->v == TH_ATPG_ZERO) || (v == TH_ATPG_ZERO && k->v == TH_ATPG_ONE))
      k->v = TH_ATPG_ONE;
    else
      k->v = TH_ATPG_ZERO;

    /* pick an input of k with value x */
    /* todo: we might wanna use a good ocst function, right now pick the first path that is x */
    for (i = 0; i < node_num_fanin(k->n); i++) {
      if (TH_ATPG_GETLINEVALUE(k->n, i) == TH_ATPG_X)
	break;
    }
    
    /* no such input found */
    if (i == node_num_fanin(k->n))
      return NIL(thatpg_assignment_t);
    
    /* if the input weight is negative, implies an invert in "digital" logic, so we flip the parity */
    if (((thnode_t *) k->n->th)->W[i] < 0)
      k->v = (k->v == TH_ATPG_ONE) ? TH_ATPG_ZERO : TH_ATPG_ONE;
    
  
    k->n = node_get_fanin(k->n, i);
  }

  //  fprintf(siserr, "backtrace: %s %s\n", node_name(k->n), (k->v == TH_ATPG_ZERO) ? "0" : "1");
  return k;
}

/* perform forward implication by making the pi assignment */
void thatpg_imply(thfault_t *f, thatpg_assignment_t *p, int flag)
{
  node_t *n, *fanout;
  lsGen gen;
  array_t *nodes;
  int i, j, k;
  thatpg_simv_t v;

  if (flag) {
    TH_SETVALUE(p->n, p->v);
    foreach_fanout(p->n, gen, fanout)
      TH_ATPG_SETLINEVALUE(fanout, node_get_fanin_index(fanout, p->n), p->v);
  }
  else {
    TH_ATPG_SETLINEVALUE(f->node, f->findex, p->v);
    TH_SETSIMCHANGED(p->n, 1);
  }
  
  if (!TH_GETSIMCHANGED(p->n))
    return;

  nodes = network_tfo(p->n, INFINITY);
  for (i = array_n(nodes) - 1; i >= 0; i--) {
    n = array_fetch(node_t *, nodes, i);
    switch (node_type(n)) {
    case PRIMARY_INPUT: 
    case PRIMARY_OUTPUT:
      break;
    case INTERNAL: 
      for (j = 0; j < node_num_fanin(n); j++) {
	if (TH_GETSIMCHANGED(node_get_fanin(n, j))) {
	  thatpg_simulate_node(n);
	  
	  /* propagate the new value to all branches, except that which contains a fault (if branch fault) */
	  v = TH_GETVALUE(n);
	  foreach_fanout(n, gen, fanout) {
	    k = node_get_fanin_index(fanout, n);
	    if (!((fanout == f->node) && (k == f->findex)))
	      TH_ATPG_SETLINEVALUE(fanout, k, v);
	  }
	  
	  /* update d-frontier if possible */
	  //thatpg_update_dfrontier(f->dfront, n);
	  break;
	}
      }
      thatpg_update_dfrontier(f->dfront, n);
      break;
    default:
      fail("illegal node type in podem imply");
    }
  }

  foreach_primary_output(p->n->network, gen, n) {
    TH_ATPG_SETLINEVALUE(n, 0, TH_GETVALUE(node_get_fanin(n, 0)));
    TH_SETVALUE(n, TH_GETVALUE(node_get_fanin(n, 0)));
  }

  array_free(nodes);

  //  thatpg_print_values(f->node->network);
}

/* update the d-frontier */
void thatpg_update_dfrontier(lsList dfront, node_t *n)
{
  thatpg_simv_t v, v1;
  int k;
  bool exist = FALSE;

  for (k = 0; k < node_num_fanin(n); k++) {
    if (TH_ATPG_GETLINEVALUE(n, k) == TH_ATPG_X) {
      exist = TRUE;
      break;
    }
  }
  
  if (!exist)
    return;
  
  /*  only if an input has d/dbar and the output is X and an input exists which is X */
  v1 = TH_GETVALUE(n);
  for (k = 0; k < node_num_fanin(n); k++) {
    v = TH_ATPG_GETLINEVALUE(n, k);
    if ((v == TH_ATPG_D || v == TH_ATPG_DBAR) && (v1 == TH_ATPG_X) && exist) {
      //      fprintf(siserr, "Adding %s to D-frontier\n", node_name(n));
      lsNewBegin(dfront, (lsGeneric) n, LS_NH);
      break;
    }
  }
}

/* simulate a node based on 5 value atpg system */
void thatpg_simulate_node(node_t *n)
{
  thnode_t *th = (thnode_t *) n->th;
  double faulty, faultfree, maxx, minx;
  int i;
  thatpg_simv_t v;

  //  fprintf(siserr, "simulating %s\n", node_name(n));

  faulty = faultfree = maxx = minx = 0.0;
  for (i = 0; i < node_num_fanin(n); i++) {
    switch (TH_ATPG_GETLINEVALUE(n, i)) {
    case TH_ATPG_ZERO: break;
    case TH_ATPG_ONE: faulty += th->W[i]; faultfree += th->W[i]; break;
    case TH_ATPG_D: faultfree += th->W[i]; break;
    case TH_ATPG_DBAR: faulty += th->W[i]; break;
    case TH_ATPG_X: 
      if (th->W[i] > 0) 
	maxx += th->W[i];
      else
	minx += th->W[i];
      break;
    default: fail("illegal node value in podem simulate node");
    }
  }
  
  /* always 0 */
  if ((faultfree + maxx < th->T) && (faulty + maxx < th->T) && (faultfree + minx < th->T) && (faulty + minx < th->T))
    v = TH_ATPG_ZERO;
  /* always 1 */
  else if ((faultfree + maxx >= th->T) && (faulty + maxx >= th->T) && (faultfree + minx >= th->T) && (faulty + minx >= th->T))
    v = TH_ATPG_ONE;
  /* d */
  else if ((faultfree >= th->T) && (faulty < th->T))
    v = TH_ATPG_D;
  /* d bar */
  else if ((faulty >= th->T) && (faultfree < th->T))
    v = TH_ATPG_DBAR;
  /* either 0/1 */
  else
    v = TH_ATPG_X;

  /* set value of node and propagate it to all branches */
  TH_SETVALUE(n, v);
}

/* returns true if error has propagated to one primary output */
bool thatpg_output_error(network_t *network)
{
  lsGen gen;
  node_t *po;
  thatpg_simv_t v;
  
  foreach_primary_output(network, gen, po) {
    v = TH_GETVALUE(po);
    if (v == TH_ATPG_D || v == TH_ATPG_DBAR) {
      lsFinish(gen);
      return TRUE;
    }
  }

  return FALSE;
}

/* return true if it is possible to find a test, false otherwise */
bool thatpg_test_possible(thfault_t *f)
{
  /* if dfrontier is empty */
  if (lsLength(f->dfront) == 0)
    return FALSE;
  
  /* if stuck at value is same as value of node */
  if ((f->value == TH_ATPG_SA0) && (TH_GETVALUE(node_get_fanin(f->node, f->findex)) == TH_ATPG_ZERO))
    return FALSE;
  if ((f->value == TH_ATPG_SA1) && (TH_GETVALUE(node_get_fanin(f->node, f->findex)) == TH_ATPG_ONE))
    return FALSE;
  
  return TRUE;
}


/* thatpg usage */
static int thatpg_usage(void)
{
  fprintf(siserr, "usage: thatpg [-d] [-c] [-n num] [-r limit] [-b limit] [-t time] [-y num] [-z num] [-o <file>] \n");
  fprintf(siserr, "    -d \t\tPerform fault dropping\n");
  fprintf(siserr, "    -c \t\tDo not perform fault collapsing\n");
  fprintf(siserr, "    -n <num> \tNumber of vectors to generate for each pass of random test generation (default is %d)\n", NUMBITS/16);
  fprintf(siserr, "    -r <limit> \tFault coverage limit during random test generation (default is 0.75)\n");
  fprintf(siserr, "    -b <limit> \tBacktrack limit during deterministic test generation (default is 1000)\n");
  fprintf(siserr, "    -t <time> \tTimeout limit for random test generation (default is 120s)\n");
  fprintf(siserr, "    -y <num> \tDelta on (default is 0).\n");
  fprintf(siserr, "    -z <num> \tDelta off (default is 0).\n");
  fprintf(siserr, "    -o <file> \tWrite test vectors to <file>\n");
  return TH_FAIL;
}

/* create a vector string from 5 value system on the network inputs */
char *thatpg_create_vector(network_t *network, int nrpi)
{
  node_t *pi;
  lsGen gen;
  char *vec;
  int i = 0;

  vec = ALLOC(char, nrpi + 1);
  memset(vec, 0, nrpi + 1);
  
  /* X can be either 0 or 1. i chose 1 arbitrarily */
  foreach_primary_input(network, gen, pi) {
    switch (TH_GETVALUE(pi)) {
    case TH_ATPG_ONE:
    case TH_ATPG_D: 
    case TH_ATPG_X: vec[i++] = '1'; break;
    case TH_ATPG_ZERO: 
    case TH_ATPG_DBAR: vec[i++] = '0'; break;
    default:
      fail("unrecognized value");
    }
  }
  
  return vec;
}
