/*=========================================================================*
 *	File    :	thatpg_util.c
 *
 *	Author  :	Pallav Gupta
 *	Created :	Wed Sep 24 17:29:15 2003
 *
 * This source code can be redistributed under the GNU Free Public Licence.
 *=========================================================================*/

/* $Id: thatpg_util.c,v 1.8 2004/03/18 20:30:41 pgupta Exp $ */

#include "sis.h"
#include "th_int.h"

/* generate all single stuck at faults, will perform fault collapsing if option is set */
int thatpg_gen_faults(thatpg_info_t *info)
{
  int i;
  node_t *n, *fanin;
  lsList faults;
  thfault_t *f;
  lsGen gen;
  st_table *fault_table;
  
  faults = lsCreate();
  foreach_node(info->network, gen, n) {
    switch (node_type(n)) {
    case PRIMARY_INPUT:
      thatpg_add_safaults(faults, n, NIL(node_t), TH_ATPG_STEM, info);
      break;
    case PRIMARY_OUTPUT:
      /* normal fault on primary output */
      break;
    case INTERNAL:
      foreach_fanin(n, i, fanin) {
	/* if input has multiple fanout, then add branch faults */
	if (node_num_fanout(fanin) > 1)
	  thatpg_add_safaults(faults, n, fanin, TH_ATPG_BRANCH, info);
	else {
	  /* otherwise a normal line fault, only if we're not doing fault collapsing */
	  if (!info->thatpg_opts->faultcollapse)
	    thatpg_add_safaults(faults, n, fanin, TH_ATPG_NORMAL, info);
	}
      }
      /* if this node fanouts out, we need to add stem fault only if we're not doing fault collapseing */
      if ((node_num_fanout(n) > 1) && !info->thatpg_opts->faultcollapse)
	thatpg_add_safaults(faults, n, NIL(node_t), TH_ATPG_STEM, info);
      break;
    default:
      fail("encountered illegal node type for fault generation");
    }
  }

  /* add all the faults to the info data structure */
  fault_table = st_init_table(st_ptrcmp, st_ptrhash);
  foreach_thfault(faults, gen, f) {
    st_insert(fault_table, (char *) f, (char *) 0);
  }
  info->fault_table = fault_table;
  lsDestroy(faults, 0);
    
  return TH_SUCCESS;
}


/* create both stuck at faults and add them to the list of faults, Assumes fault is a valid handle */
void thatpg_add_safaults(lsList faults, node_t *n, node_t *fanin, thfault_type_t type, thatpg_info_t *info)
{
  thfault_t *f;

  f = thfault_init(n, fanin, TH_ATPG_SA0, type, info);
  lsNewEnd(faults, (lsGeneric) f, 0);
  f = thfault_init(n, fanin, TH_ATPG_SA1, type, info);
  lsNewEnd(faults, (lsGeneric) f, 0);
}

/* initialize the thatpg_info_t data structure */
thatpg_info_t *thatpg_info_init(network_t *network)
{
  thatpg_info_t *info;
  lsGen gen;
  node_t *n;

  info = ALLOC(thatpg_info_t, 1);

  info->thatpg_opts = NIL(thatpg_options_t);
  info->seq = network_num_latch(network) ? TRUE : FALSE;

  info->tests = NIL(st_table);
  info->fault_table = NIL(st_table);

  info->simv = NIL(array_t);
  info->simr = NIL(array_t);
  info->simf = NIL(array_t);
  info->nv = 0;

  info->npi = network_num_pi(network);
  info->npo = network_num_po(network);
  info->nlatch = network_num_latch(network);
  info->network = network;

  info->nrpo = 0;
  foreach_primary_output(network, gen, n)
    if (network_is_real_po(network, n))
      info->nrpo++;

  info->nrpi = 0;
  foreach_primary_input(network, gen, n)
    if (network_is_real_pi(network, n))
      info->nrpi++;

  info->ilpmodel1_file = util_tempnam(NIL(char), "SIS");
  info->ilpmodel2_file = util_tempnam(NIL(char), "SIS");
  info->ilpdata_file = util_tempnam(NIL(char), "SIS");
  info->ilpresult_file = util_tempnam(NIL(char), "SIS");

  /* add other stuff as necessary */
  return info;
}

/* free the thatgp_info_t data structure */
void thatpg_info_free(thatpg_info_t *info)
{
  st_generator *gen;
  thfault_t *f;
  char *v;

  if (info == NIL(thatpg_info_t))
    return;

  thatpg_options_free(info->thatpg_opts);

  /* free the test table */
  if (info->tests != NIL(st_table)) {
    st_foreach_item(info->tests, gen, (char **) &v, (char **) 0)
      FREE(v);
    st_free_table(info->tests);
  }
  
  /* free the fault table */
  if (info->fault_table != NIL(st_table)) {
    st_foreach_item(info->fault_table, gen, (char **) &f, (char **) 0)
      thfault_free(f);
    st_free_table(info->fault_table);
  }

  /* free the arrays */
  if (info->simv != NIL(array_t))
    array_bigint_free(info->simv, info->nrpi);
  if (info->simr != NIL(array_t))
    array_bigint_free(info->simr, info->nrpo);
  if (info->simf != NIL(array_t))
    array_bigint_free(info->simf, info->nrpo);
  
  if (info->ilpmodel1_file != NIL(char)) {
    unlink(info->ilpmodel1_file);
    FREE(info->ilpmodel1_file);
  }
  if (info->ilpmodel2_file != NIL(char)) {
    unlink(info->ilpmodel2_file);
    FREE(info->ilpmodel2_file);
  }
  if (info->ilpdata_file != NIL(char)) {
    unlink(info->ilpdata_file);
    FREE(info->ilpdata_file);
  }
  if (info->ilpresult_file != NIL(char)) {
    unlink(info->ilpresult_file);
    FREE(info->ilpresult_file);
  }
  
  FREE(info);
  /* add other stuff as necessary */
}

/* initialize the thatpg_options_t data structure */
thatpg_options_t *thatpg_options_init(int argc, char **argv)
{
  int c, err = 0;
  char *filename;
  thatpg_options_t *opt;

  opt = ALLOC(thatpg_options_t, 1);

  /* default values */
  opt->verb = 0;
  opt->ifile = NIL(char);
  opt->ofile = NIL(char);
  opt->fpi = NIL(FILE);
  opt->fpo = NIL(FILE);
  opt->help = FALSE;
  opt->faultdrop = FALSE;
  opt->faultcollapse = TRUE;
  opt->fcoverage_limit = 0.75;
  opt->rand_timeout = 120; /* seconds */
  opt->backtrack_limit = 1000;
  opt->don = opt->doff = 0;
  opt->nvec = NUMBITS/16;

  /* parse options */
  util_getopt_reset();
  while ((c = util_getopt(argc, argv, "i:o:v:r:t:y:z:dchb:n:")) != EOF) {
    if (err)
      break;

    switch (c) {
    case 'i':
      opt->fpi = com_open_file(util_optarg, "r", &filename, /* silent */ 0);
      opt->ifile = filename;
      if (opt->fpi == NULL) {
	FREE(filename);
	fprintf(siserr, "Unable to open file '%s' to read\n", util_optarg);	
	err = 1; 
      }
      break;
    case 'o':
      opt->fpo = com_open_file(util_optarg, "w", &filename, /* silent */ 0);
      opt->ofile = filename;
      if (opt->fpo == NULL) {
	FREE(filename);
	fprintf(siserr, "Unable to open file '%s' to write\n", util_optarg); 
	err = 1;
      }
      break;
    case 'r':
      opt->fcoverage_limit = atof(util_optarg);
      if ((opt->fcoverage_limit < 0.0) && (opt->fcoverage_limit > 1.0)) {
	fprintf(siserr, "Fault coverage limit must be between 0 - 1.0\n");
	err = 1;
      }
      break;
    case 't':
      opt->rand_timeout = atoi(util_optarg);
      if (opt->rand_timeout < 0) {
	fprintf(siserr, "Timeout must be greater than 0\n");
	err = 1;
      }
      break;
    case 'b':
      opt->backtrack_limit = atoi(util_optarg);
      if (opt->backtrack_limit < 0) {
	fprintf(siserr, "Backtrack limit must be grater than 0\n");
	err = 1;
      }
      break;
    case 'n':
      opt->nvec = atoi(util_optarg);
      if (opt->nvec < 1 || opt->nvec > NUMBITS) {
	fprintf(siserr, "Number of vectors for each phase of fault simulation must be between 1 - %d\n", NUMBITS);
	err = 1;
      }
      break;
    case 'y': opt->don = atof(util_optarg); break;
    case 'z': opt->doff = atof(util_optarg); break;
    case 'v': opt->verb = atoi(util_optarg); break;
    case 'd': opt->faultdrop = TRUE; break;
    case 'c': opt->faultcollapse = FALSE; break;
    case 'h': opt->help = TRUE; break;
    default: opt->help = TRUE; break;
    }
  }

  /* consistency check */
  if (!(opt->don >= -(opt->doff))) {
    fprintf(siserr, "Parameters don/doff are inconsistent. don must be >= -doff.\n");
    err = 1;
  }

  /* add other stuff as necessary */
  if (err) {
    FREE(opt);
    return NULL;
  }
  else
    return opt;
}

/* free the thatpg_options_t data sturcture */
void thatpg_options_free(thatpg_options_t *opt)
{
  if (opt == NIL(thatpg_options_t))
    return;

  if (opt->ifile != NIL(char))
     FREE(opt->ifile);
   if (opt->ofile != NIL(char))
     FREE(opt->ofile);
   if (opt->fpi != NIL(FILE))
     fclose(opt->fpi);
   if (opt->fpo != NIL(FILE))
     fclose(opt->fpo);

   FREE(opt);
}

/* initialize a thfault_t data structure */
thfault_t *thfault_init(node_t *n, node_t *fanin, thfault_value_t value, thfault_type_t type, thatpg_info_t *info)
{
  thfault_t *f;

  f = ALLOC(thfault_t, 1);

  f->node = n;
  f->fanin = fanin;

  if (f->fanin == NIL(node_t))
    f->findex = -1;
  else
    f->findex = node_get_fanin_index(n, fanin);

  f->value = value;
  f->type = type;
  f->status = TH_ATPG_UNTESTED;
  f->dfront = lsCreate();
  f->backtracks = 0;

  /* add other stuff as necessary */
  return f;
}

/* free a thfault_t data structure */
void thfault_free(thfault_t *f)
{
  if (f == NIL(thfault_t))
    return;

  lsDestroy(f->dfront, NIL(void));

  FREE(f);
}

/* display the faults list */
void thatpg_print_all_faults(FILE *fp, thatpg_info_t *info)
{
  st_generator *gen;
  thfault_t *f;

  TH_ASSERT(info->fault_table != NIL(st_table));

  st_foreach_item(info->fault_table, gen, (char **) &f, (char **) 0)
    thatpg_print_fault(fp, f);
}

/* print a single fault */
void thatpg_print_fault(FILE *fp, thfault_t *f)
{
  char *sa[2] = {"SA0", "SA1"};
  char *type[3] = {"stem", "branch", "line"};
  char *s, *t, *p;

  TH_ASSERT(f != NIL(thfault_t));

  s = (f->value == TH_ATPG_SA0) ? sa[0] : sa[1];
  p = (f->findex != -1) ? node_name(f->fanin) : node_name(f->node);
  
  if (f->type == TH_ATPG_STEM)
    t = type[0];
  else if (f->type == TH_ATPG_BRANCH)
    t = type[1];
  else 
    t = type[2];
  
  fprintf(fp, "%s fault on %s of %s on node %s\n", s, t, p, node_name(f->node));
}


/* display the test vector list */
void thatpg_print_test_vectors(thatpg_info_t *info)
{
  st_generator *gen;
  char *t;
  FILE *fp;

  fp = info->thatpg_opts->fpo != NIL(FILE) ? info->thatpg_opts->fpo : sisout;
  TH_ASSERT(info->tests != NIL(st_table));

  fprintf(fp, "Threshold ATPG test vectors:\n");
  st_foreach_item(info->tests, gen, (char **) &t, (char **) 0)
    fprintf(fp, "%s\n", t);
}

/* read in test vectors from file and put in a hash table each vector must 
   exists in one line with no spaces between the bits */
int thatpg_read_test_vectors(thatpg_info_t *info)
{
  st_table *tests;
  FILE *fp;
  char *linep, *buf;
  int linen;

  linen = 0;
  linep = NIL(char);
  fp = info->thatpg_opts->fpi;
  TH_ASSERT(fp != NIL(FILE));

  /* store the vectors as strings */
  tests = st_init_table(strcmp, st_strhash);

  /* read each vector on a single line */
  while (read_line(fp, &linep, &linen) != EOF) {
    if (!st_lookup(tests, linep, (char **) 0)) {
      if (strlen(linep) != info->nrpi) {
	fprintf(siserr, "Vector size does not contain %d bits\n", info->nrpi);
	FREE(linep);
	st_free_table(tests);
	return TH_FAIL;
      }
	
      buf = ALLOC(char, info->nrpi + 1);
      strcpy(buf, linep);
      st_insert(tests, buf, (char *) 0);
    }
    /* duplicate vector */
    else 
      fprintf(siserr, "Ignoring duplicate vector %s\n", linep);
  }

  info->tests = tests;
  FREE(linep);
  return TH_SUCCESS;
}


/* dump statitics about atpg */
void thatpg_dump_statistics(thatpg_info_t *info)
{
  st_generator *gen;
  thfault_t *f;
  int detected, redundant;
  FILE *fp;

  fp = info->thatpg_opts->fpo != NIL(FILE) ? info->thatpg_opts->fpo : sisout;

  /* count all detected faults */
  detected = redundant = 0;
  st_foreach_item(info->fault_table, gen, (char **) &f, 0) {
    if (f->status == TH_ATPG_TESTED)
      detected++;
    else if (f->status == TH_ATPG_REDUNDANT)
      redundant++;
  }
  
  /* dump useful statistics */
  fprintf(fp, "Threshold ATPG statistics:\n");
  fprintf(fp, "# of faults: \t %d\n", st_count(info->fault_table));
  fprintf(fp, "# detected:  \t %d\n", detected);
  fprintf(fp, "# undetected:\t %d\n", st_count(info->fault_table) - detected);
  fprintf(fp, "# redundant: \t %d\n", redundant);
  if (info->tests != NIL(st_table))
    fprintf(fp, "# of vectors:\t %d\n", st_count(info->tests));
  fprintf(fp, "coverage: \t %lf\n", (double) (detected + redundant)/st_count(info->fault_table) * 100);
  
  /* print additional network information */
  thprint_stats(info->network, fp, NIL(array_t));
}
