/*=========================================================================*
 *	File    : thfsim.c
 *
 *	Author  : Pallav Gupta
 *	Created : Tue Sep 23 20:59:34 2003
 *
 * This source code can be redistributed under the GNU Free Public Licence.
 *=========================================================================*/

/* $Id: thfsim.c,v 1.6 2004/01/16 03:51:58 pgupta Exp $ */

#include "sis.h"
#include "th_int.h"

static int thfsim_usage ARGS(());
static void inject_fault ARGS((thfault_t *, thnode_t *, int));
static void remove_fault ARGS((thfault_t *, thnode_t *, int));

/* threshold fault simulation */
int com_thfsim(network_t **network, int argc, char **argv)
{
  thatpg_info_t *info;
  thatpg_options_t *opt;
  FILE *fp;

   /* check if threshold network */
  if (is_threshold_network(*network) != TH_SUCCESS)
    return TH_FAIL;

  /* initialize data structures */
  info = thatpg_info_init(*network);
  opt = thatpg_options_init(argc, argv);
  info->thatpg_opts = opt;

  if (opt == NULL) {
    thatpg_info_free(info);
    return TH_FAIL;
  }
  else if (opt->help) {
    thatpg_info_free(info);
    return thfsim_usage();
  }
  /* TODO: FSM test generation */
  else if (info->seq) {
    fprintf(siserr, "We don't support FSM fault simulation just yet\n");
    thatpg_info_free(info);
    return TH_FAIL;
  }

  if (opt->fpi == NIL(FILE)) {
    fprintf(siserr, "Must specify input file to read vectors from\n");
    thatpg_info_free(info);
    return TH_FAIL;
  }

  /* Generate all the faults */
  if (thatpg_gen_faults(info) != TH_SUCCESS) {
    fprintf(siserr, "Error in generating faults\n");
    thatpg_info_free(info);
    return TH_FAIL;
  }

  fp = opt->fpo != NIL(FILE) ? opt->fpo : sisout;

  if (info->thatpg_opts->verb > 2)
    thatpg_print_all_faults(fp, info);

  /* Read in all the test vectors from file */
  TH_ASSERT((info->simv = th_read_test_vectors(info->network, opt->fpi, &(info->nv))) != NIL(array_t));
  
  /* Simulate all the test vectors to get the fault-free values */
  TH_ASSERT((info->simr = thfsim_simulate_vectors(info)) != NIL(array_t));

  /* Let us simulate all the faults now */
  thfsim_simulate_all_faults(info);
  
  /* dump statistics about this fault simulation run */
  thatpg_dump_statistics(info);

  /* free and cleanup */
  thatpg_info_free(info);
  return TH_SUCCESS;
}



/* simualte all the faults */
void thfsim_simulate_all_faults(thatpg_info_t *info) 
{
  st_generator *gen;
  thfault_t *f;
  FILE *fp;

  fp = info->thatpg_opts->fpo != NIL(FILE) ? info->thatpg_opts->fpo : sisout;

  /* simulate each fault */
  st_foreach_item(info->fault_table, gen, (char **) &f, (char **) 0) {
    if (info->thatpg_opts->verb > 1)
      thatpg_print_fault(fp, f);

    thfsim_simulate_fault(info, f);
  
    if (info->thatpg_opts->verb > 1) {
      if (f->status == TH_ATPG_TESTED)
	fprintf(fp, "status = DETECTED\n");
      else
	fprintf(fp, "status = NOT\n");
    }
  }
}

/* simulat ea single fault */		   
void thfsim_simulate_fault(thatpg_info_t *info, thfault_t *f)
{
  /* inject this fault in the network */
  thfsim_inject_fault(f);

  /* do fault dropping? i.e. as soon as one bigint word detects the fault, don't bother proceeding anymore*/
  if (info->thatpg_opts->faultdrop) 
    f->status = thfsim_sim_and_compare_vectors(info);
  else {
    /* simulate all the test vectors */
    info->simf = thfsim_simulate_vectors(info);
    /* determine if there exists one vector which has different outputs */
    f->status = compare_sim_results(info->simr, info->simf, info->nrpo, info->nv) ? TH_ATPG_TESTED : TH_ATPG_UNTESTED;
    array_bigint_free(info->simf, info->nrpo);
    info->simf = NIL(array_t);
  }

  /* remove the fault from the network */
  thfsim_remove_fault(f);
}

/* simulate and compare vectors one bigint at a time, this way we implicity implement fault dropping 
 because as soon as we find a vector that is different, we don't bother proceeding further */
int thfsim_sim_and_compare_vectors(thatpg_info_t *info)
{
  BIGINT **res, **exp;
  int i, j, n, bits, status;

  for (i = 0, n = info->nv; i < array_n(info->simv); i++, n -= NUMBITS) {
    res = thsim_comb_thnetwork(info->network, array_fetch(BIGINT **, info->simv, i), MAXENTRIES);
    exp = array_fetch(BIGINT **, info->simr, i);

    bits = (n >= NUMBITS) ? MAXBITINDEX : (n - 1);

    for (j = 0; j < info->nrpo; j++) {
      TH_ASSERT((status = bigint_compare(exp[j], res[j], bits)) != BIGINT_ERROR);
      if (status == BIGINT_DIFF) {
	bigint_array_free(res, info->nrpo);
	return TH_ATPG_TESTED;
      }
    }
    bigint_array_free(res, info->nrpo);
  }

  return TH_ATPG_UNTESTED;
}


/* simulate a set of vectors and return an array contain the results */
array_t *thfsim_simulate_vectors(thatpg_info_t *info) 
{
  array_t *a;
  int i, n;
  FILE *fp;

  a = array_alloc(BIGINT **, 0);
  for (i = 0; i < array_n(info->simv); i++)
    array_insert_last(BIGINT **, a, thsim_comb_thnetwork(info->network, array_fetch(BIGINT **, info->simv, i), MAXENTRIES));

  if (info->thatpg_opts->verb > 2) {
    fp = info->thatpg_opts->fpo != NIL(FILE) ? info->thatpg_opts->fpo : sisout;
    n = info->nv;
    for (i = 0; i < array_n(a); i++) {
      if (n >= NUMBITS)
	print_simresults(info->network, fp, array_fetch(BIGINT **, a, i), NUMBITS);
      else
	print_simresults(info->network, fp, array_fetch(BIGINT **, a, i), n);
      n -= NUMBITS;
    }
  }

  return a;
}


/* inject fault on node */
void thfsim_inject_fault(thfault_t *f)
{
  thnode_t *th;
  lsGen gen;
  node_t *fanout;

  TH_ASSERT(f != NIL(thfault_t));
  TH_ASSERT(f->status != TH_ATPG_TESTED);

  /* fault on an internal node */
  if (f->findex != -1) {
    /* branch and normal faults can be dealt with ordinarily */
    thnode_check(f->node);
    th = (thnode_t *) f->node->th;
    inject_fault(f, th, f->findex);
  }
  else {
    /* if its a stem fault, we need to propagate its effect on
       all branches */
    if (f->type == TH_ATPG_STEM) {
      foreach_fanout(f->node, gen, fanout) {
	thnode_check(fanout);
	th = (thnode_t *) fanout->th;
	inject_fault(f, th, node_get_fanin_index(fanout, f->node));
      }
    }
    else
      TH_ASSERT(1); /* should not be reached */
  }
}

/* remove fault on node */
void thfsim_remove_fault(thfault_t *f)
{
  thnode_t *th;
  lsGen gen;
  node_t *fanout;

  TH_ASSERT(f != NIL(thfault_t));
    
  if (f->findex != -1) {
    /* branch and normal faults can be dealt with ordinarily */
    thnode_check(f->node);
    th = (thnode_t *) f->node->th;
    remove_fault(f, th, f->findex);
  }
  else {
    if (f->type == TH_ATPG_STEM) {
      foreach_fanout(f->node, gen, fanout) {
	thnode_check(fanout);
	th = (thnode_t *) fanout->th;
	remove_fault(f, th, node_get_fanin_index(fanout, f->node));
      }
    }
    else
      TH_ASSERT(1); /* should not be reached */
  }
}

/* thfsim usage */
static int thfsim_usage(void)
{
  fprintf(siserr, "usage: thfsim [-d] [-c] [-o file] -i <file> \n");
  fprintf(siserr, "    -d \t\tPerform fault dropping\n");
  fprintf(siserr, "    -c \t\tDo not perform fault collapsing\n");
  fprintf(siserr, "    -i <file> \tRead test vectors from <file>\n");
  fprintf(siserr, "    -o <file> \tWrite all detected faults to <file>\n");
  return TH_FAIL;
}

/* inject the fault in the node */
static void inject_fault(thfault_t *f, thnode_t *t, int i)
{
  /* if SA1 fault, need to reduce threshold */
  if (f->value == TH_ATPG_SA1)
    t->T -= t->W[i];
  /* in either case, the weight goes to 0.0 */
  t->scratch = t->W[i];
  t->W[i] = 0.0;
}

/* remove the fault in the node */
static void remove_fault(thfault_t *f, thnode_t *t, int i)
{
  /* restore threshold */
  if (f->value == TH_ATPG_SA1)
    t->T += t->scratch;
  /* restore weight */
  t->W[i] = t->scratch;
  t->scratch = 0.0;
}
