/*=========================================================================*
 *      File    : thsyn_util.c
 *
 *      Author  : Rui Zhang
 *      Created : Tue Mar 23 17:22:47 2004
 *
 * This source code can be redistributed under the GNU Free Public Licence.
 *=========================================================================*/

/* $Id: thsyn_util.c,v 1.6 2005/06/03 18:01:48 sis Exp $ */

#include "sis.h"
#include "th_int.h"

static int node_compare(node_t **n1, node_t **n2);

/************************************************************************ 
 * Given an array of nodes, this function sort this array of nodes      *
 * in decreasing order with respect to the number of cubes of each node *
 ************************************************************************/
void thsyn_sort_nodes(array_t *nodes)
{
  /* sort the array */
  array_sort(nodes, node_compare);
}

/*********************************************************************** 
 * This function itself is used as a parameter in function: array_sort *
 * It compares two nodes by the number of cubes it contains            *
 * returns -1 if n1 is greater, 0 otherwise                            *
 ***********************************************************************/
static int node_compare(node_t **n1, node_t **n2)
{
  int c1, c2;

  /* get number of cubes */
  c1 = node_num_cube(*n1);
  c2 = node_num_cube(*n2);
  /* return appropriate status */
  if (c1 > c2)
    return -1;
  else if (c1 == c2)
    return 0;
  else
    return 1;
}

/***************************************************************************** 
 * This function checks if all the fanins of a given node are primary inputs.* 
 * It takes parameter: node_t * which is a node in sis network               *
 * If all fanin are primary inputs, return 1, otherwise 0                    *
 *****************************************************************************/
int thsyn_if_all_PI(node_t *n)
{
  int i, status = 1;
  node_t *fanin;
  
  foreach_fanin(n, i, fanin) {
    if (node_type(fanin) != PRIMARY_INPUT) {
      status = 0;
      break;
    }
  }
  return status;
}

/*****************************************************************************
 * This function create a file in AMPL syntax, which describes the model of  *
 * lp problem. It takes parameter: opt.  AMPL will read the model from the   *
 * created file and solve it later together with the corresponding data file.*
 *****************************************************************************/
void thsyn_write_ampl_model(thsyn_options_t *opt)
{
  FILE *fp;
  char *filename;
  
  filename = ALLOC(char, TH_MAXWORD);
  strcpy(filename, opt->filename);
  strcat(filename, ".mod");

  /* open file handle or abort */
  if ((fp = fopen(filename, "w")) == NULL)
    fail("thsyn_write_ampl_model: unable to open file for writing");

  /* define the parameters and variables for the model */
  fprintf(fp, "# All the inputs \nset X;\n\n");
  fprintf(fp, "# number of on-set cubes \nset ON;\n\n");
  fprintf(fp, "# number of off-set cubes \nset OFF;\n\n");
  fprintf(fp, "# parameter xon [j] \nparam xon {i in ON, j in X} binary;\n\n");
  fprintf(fp, "# parameter xoff [j] \nparam xoff {k in OFF, j in X} binary;\n\n");
  fprintf(fp, "# decision variables w[j] \nvar w {j in X} >= 0 integer;\n\n");
  fprintf(fp, "# decision variables threshold t \nvar t >= 0 integer;\n\n");

  /* write objective function */
  fprintf(fp, "minimize cost: (sum {j in X} w[j]) + t;\n\n");

  /* write constraint inequalities */
  /* on-set inequalities */
  fprintf(fp, "subject to onineq {i in ON}: \n\tsum {j in X} (w[j] * xon[i,j]) >= t + (%d);\n\n", opt->delta_on);
  /* off-set inequalities */
  fprintf(fp, "subject to offineq {k in OFF}: \n\tsum {j in X} (w[j] * xoff[k,j]) <= t - (%d);\n", opt->delta_off); 

  fclose(fp);
  FREE(filename);
}

/*********************************************************
 * write the command file to call AMPL by system command *
 *********************************************************/
void thsyn_write_ampl_commands(thsyn_options_t *opt)
{
  FILE *fp;
  
  /* open file handle or abort */
  if ((fp = fopen("thsyn_ampl_commands", "w")) == NULL)
    fail("thsyn_write_ampl_commands: unable to open file handle for writing");
  
  /*create the command to read in the AMPL model file */
  fprintf(fp, "model %s.mod;\n", opt->filename);
  /* write the command to read in the AMPL model file */
  fprintf(fp, "data %s.data;\n", opt->filename);
  /* write the command to set AMPL options */
  fprintf(fp, "option solver cplex;\n");
  fprintf(fp, "option solver_msg 0;\n");
  /* solve the AMPL problem */
  fprintf(fp, "solve;\n");
  /* display the results */
  fprintf(fp, "print t, {j in X} w[j] > thsyn_ampl.out;\n");
  fclose(fp);
}

/*****************************************
 * write the ampl data file for the node *
 *****************************************/
void thsyn_write_ampl_data(node_t *node, thsyn_options_t *opt)
{
  int num, index, unate;
  char *filename;
  node_t *inv;
  FILE *fp;

  filename = ALLOC(char, TH_MAXWORD);
  strcpy(filename, opt->filename);
  strcat(filename, ".data");
  
  /* open file handle or abort */
  if ((fp = fopen(filename, "w")) == NULL)
    fail("thsyn_write_ampl_data: unable to open file for writing");
  
  /* write the data corresponding to the model, write the number of fanins */
  num = node_num_fanin(node);
  fprintf(fp, "set X :=");
  for (index = 1; index <= num; index++)
    fprintf(fp, " %d", index);
  fprintf(fp, ";\n");

  /* write the number of on-set cubes */
  num = node_num_cube(node);
  fprintf(fp, "set ON :=");
  for (index = 1; index <= num; index++)
    fprintf(fp, " %d", index);
  fprintf(fp, ";\n");

  /* compute invert of node to get OFF set */
  inv = node_not(node);

  /* write the number of off-set cubes */
  num = node_num_cube(inv);
  fprintf(fp, "set OFF :=");
  for (index = 1; index <= num; index++)
    fprintf(fp, " %d", index);
  fprintf(fp, ";\n");

  unate = thsyn_check_unate(node);

  /* write the on-set cubes contents */
  thsyn_write_inequalities(fp, node, unate, 1);
  /* write the off-set cubes contents */
  thsyn_write_inequalities(fp, inv, unate, 0);

  node_free(inv);
  FREE(filename);
  fclose(fp);

}

/*************************************************************************************
 * Called by thsyn_write_ampl_data only to dump inequalities data into the data file *
 * for AMPL to solve. The inequalities data correspond to the ON set and OFF set of  *
 * the logic function respectively, depending on the set_type in the parameter.      *
 *************************************************************************************/
void thsyn_write_inequalities(FILE *fp, node_t *node, int unate, int set_type)
{
  int i, j, num_cubes, num_fanin;
  node_cube_t c;
  node_literal_t literal;
  input_phase_t cube_choice, type;
  node_t *fanin;
  
  type = (set_type) /* ON */ ? NEG_UNATE : POS_UNATE;
  
  if (set_type) /* ON */
    fprintf(fp, "param xon :"); 
  else /* OFF */
    fprintf(fp, "param xoff :");  
  
  num_fanin = node_num_fanin(node);
  for (i = 1; i <= num_fanin; i++)
    fprintf(fp, " %d", i);
  fprintf(fp, " :=\n");

  num_cubes = node_num_cube(node);
  for (i = 0; i < num_cubes; i++) {
    fprintf(fp, "%d ", i+1);
    c = node_get_cube(node, i);
    
    /* write out each inequality */
    foreach_fanin(node, j, fanin) {
      cube_choice = node_input_phase(node, fanin);
      literal = node_get_literal(c, j);
      
      if (literal == ONE) {
	if ((unate == NEG_UNATE) && (cube_choice == type))
	  fprintf(fp, " 0");
	else 
	  fprintf(fp, " 1");
      }
      else if (literal == ZERO) {
	if ((unate == NEG_UNATE) && (cube_choice == type)) 
	  fprintf(fp, " 1");
	else
	  fprintf(fp, " 0");
      }
      else if (literal == TWO) {
	if (!set_type) /* OFF */
	  fprintf(fp, " 1");
	else
	  fprintf(fp, " 0");
      }
      else
	fail("bad literal");
    }
    fprintf(fp, "\n");
  }
  fprintf(fp, ";\n\n");
}

/* ***********************************************************************
 *  This function checks the unateness of the node                       *
 *  it takes parameter: node_t * which is a node in sis network          *
 *  returns:                                                             *
 *  POS_UNATE: if node is positive in all variables                      *
 *  NEG_UNATE: if node is unate, but some variables are in negative form *
 *  BINATE:    if node is binate in any variable                         *
 *************************************************************************/
int thsyn_check_unate(node_t *node)
{
  int j, pos = 1;
  node_t *fanin;
  
  foreach_fanin(node, j, fanin) {
    switch (node_input_phase(node, fanin)) {
    case POS_UNATE: break;
    case NEG_UNATE: pos = 0; break;
    case BINATE: return BINATE;
    default: fail("phase unknown\n");
    }
  }
  if (pos)
    return POS_UNATE;
  else
    return NEG_UNATE;
}

/******************************************* 
 * initialize thsyn options data structure *
 *******************************************/
thsyn_options_t *thsyn_options_init(int argc, char **argv)
{
  int c, error;
  thsyn_options_t *opt;
  
  opt = ALLOC(thsyn_options_t, 1);

  /* default values */
  opt->max_fanin = 6; 
  opt->delta_on = 0;
  opt->delta_off = 1;
  opt->filename = util_strsav("thsyn_ampl");
  opt->help = FALSE;

  /* parse options */
  util_getopt_reset();
  while ((c = util_getopt(argc, argv, "c:n:f:o:h")) != EOF) {
    switch (c) {
    case 'c': opt->max_fanin = atoi(util_optarg); break;
    case 'n': opt->delta_on = atoi(util_optarg); break;
    case 'f': opt->delta_off = atoi(util_optarg); break;
    case 'o': opt->filename = util_strsav(util_optarg); break;
    case 'h': opt->help = TRUE; break;
    default: opt->help = TRUE; break;
    }
  }

  /* check if ampl/cplex are present */
  error = 0;
  if (util_path_search(AMPL) == NIL(char)) {
    fprintf(siserr, "AMPL not found. Please download from www.ampl.com and install\n");
    error = 1;
  }
  if (util_path_search(ILP_SOLVER) == NIL(char)) {
    fprintf(siserr, "CPLEX not found. Please download from www.cplex.com and insall\n");
    error = 1;
  }
  
  if (opt->max_fanin < 1 || !(opt->delta_on >= (-opt->delta_off)) || error) {
    thsyn_options_free(opt);
    return NIL(thsyn_options_t);
  }
  
  return opt;
}

int thsyn_interconnect(network_t *network)
{ 
  int connect = 0;
  node_t *n;
  lsGen gen;
  
  foreach_node(network, gen, n)
    if (node_type(n) == INTERNAL) 
      connect += node_num_fanin(n); 
	
  return connect; 
}

/************************************
 * for experiments only, collecting *
 *  area and level data             *     
 ************************************/
double thsyn_area(network_t *network)
{
  double area;
  node_t *n;
  lsGen gen;

  area = 0.0;
  foreach_node(network, gen, n)
    if (node_type(n) == INTERNAL) 
      area += thnode_calc_area(n, NULL);

  return area;
}

int thsyn_levels(network_t *network)
{
  int levels;
  node_t *n;
  lsGen gen;
  array_t *node_array;

  levels = 0;
  node_array = array_alloc(node_t *, 0);
  foreach_primary_output(network, gen, n) {
    array_insert_last(node_t *, node_array, n);
  }
  levels = speed_print_level(network, node_array, 0.5, 0, 0) + 1;
  
  return levels;
}


/*******************************
 * free options data structure *
 *******************************/
void thsyn_options_free(thsyn_options_t *opt)
{
  if (opt == NIL(thsyn_options_t))
    return;

  if (opt->filename != NIL(char))
    FREE(opt->filename);

  FREE(opt);
}
